from __future__ import annotations

import os
from pathlib import Path

from pydantic import ValidationError

from kanbus import config_loader
from kanbus.models import RealtimeConfig


def test_parse_bool_env_supports_common_values(monkeypatch) -> None:
    monkeypatch.setenv("K_BOOL", "yes")
    assert config_loader._parse_bool_env("K_BOOL") is True
    monkeypatch.setenv("K_BOOL", "OFF")
    assert config_loader._parse_bool_env("K_BOOL") is False
    monkeypatch.setenv("K_BOOL", "maybe")
    assert config_loader._parse_bool_env("K_BOOL") is None


def test_parse_int_env_handles_invalid_values(monkeypatch) -> None:
    monkeypatch.setenv("K_INT", "42")
    assert config_loader._parse_int_env("K_INT") == 42
    monkeypatch.setenv("K_INT", "nope")
    assert config_loader._parse_int_env("K_INT") is None


def test_load_dotenv_sets_values_without_overwriting_existing(
    tmp_path: Path, monkeypatch
) -> None:
    dotenv = tmp_path / ".env"
    dotenv.write_text(
        "\n".join(
            [
                "# comment",
                "PLAIN=one",
                'export QUOTED="two"',
                "KEEP=from_file",
                "INVALID_LINE",
            ]
        )
        + "\n",
        encoding="utf-8",
    )
    monkeypatch.setenv("KEEP", "already")

    config_loader._load_dotenv(dotenv)

    assert os.environ.get("PLAIN") == "one"
    assert os.environ.get("QUOTED") == "two"
    assert os.environ.get("KEEP") == "already"


def test_normalize_virtual_projects_converts_list_to_dict() -> None:
    data = {"virtual_projects": []}
    config_loader._normalize_virtual_projects(data)
    assert data["virtual_projects"] == {}


def test_reject_legacy_fields_promotes_external_projects() -> None:
    data = {"external_projects": {"team": {"project_directory": "project"}}}
    config_loader._reject_legacy_fields(data)
    assert "external_projects" not in data
    assert "virtual_projects" in data


def test_validate_sort_rule_captures_invalid_shapes() -> None:
    errors: list[str] = []
    config_loader._validate_sort_rule("sort_order.test", "bad-preset", errors)
    config_loader._validate_sort_rule(
        "sort_order.test",
        [{"field": "bad-field", "direction": "up"}],
        errors,
    )
    config_loader._validate_sort_rule("sort_order.test", [{}], errors)

    joined = " | ".join(errors)
    assert "invalid preset" in joined
    assert "invalid field" in joined
    assert "invalid direction" in joined
    assert "missing 'field'" in joined
    assert "missing 'direction'" in joined


def test_load_configuration_data_handles_empty_and_invalid_mapping(
    tmp_path: Path,
) -> None:
    empty_cfg = tmp_path / "empty.yml"
    empty_cfg.write_text("", encoding="utf-8")
    assert config_loader._load_configuration_data(empty_cfg) == {}

    non_mapping_cfg = tmp_path / "list.yml"
    non_mapping_cfg.write_text("- one\n- two\n", encoding="utf-8")
    try:
        config_loader._load_configuration_data(non_mapping_cfg)
    except config_loader.ConfigurationError as error:
        assert "configuration must be a mapping" in str(error)
    else:
        raise AssertionError("expected ConfigurationError")


def test_load_override_configuration_handles_missing_invalid_and_non_mapping(
    tmp_path: Path,
) -> None:
    missing = tmp_path / "missing.override.yml"
    assert config_loader._load_override_configuration(missing) == {}

    invalid = tmp_path / "invalid.override.yml"
    invalid.write_text("x: [\n", encoding="utf-8")
    try:
        config_loader._load_override_configuration(invalid)
    except config_loader.ConfigurationError as error:
        assert "override configuration is invalid" in str(error)
    else:
        raise AssertionError("expected ConfigurationError")

    non_mapping = tmp_path / "list.override.yml"
    non_mapping.write_text("- one\n", encoding="utf-8")
    try:
        config_loader._load_override_configuration(non_mapping)
    except config_loader.ConfigurationError as error:
        assert "override configuration must be a mapping" in str(error)
    else:
        raise AssertionError("expected ConfigurationError")


def test_validate_canonical_config_overrides_rejects_hierarchy_override(
    tmp_path: Path,
) -> None:
    cfg = tmp_path / "kanbus.yml"
    cfg.write_text("project_key: k\n", encoding="utf-8")
    try:
        config_loader._validate_canonical_config_overrides(
            cfg, {"hierarchy": ["epic", "task"]}
        )
    except config_loader.ConfigurationError as error:
        assert "hierarchy is fixed" in str(error)
    else:
        raise AssertionError("expected ConfigurationError")


def test_apply_environment_overrides_populates_nested_realtime_overlay(
    monkeypatch,
) -> None:
    monkeypatch.setenv("KANBUS_REALTIME_TRANSPORT", "mqtt")
    monkeypatch.setenv("KANBUS_REALTIME_BROKER", "mqtt://localhost:1883")
    monkeypatch.setenv("KANBUS_REALTIME_AUTOSTART", "true")
    monkeypatch.setenv("KANBUS_REALTIME_KEEPALIVE", "false")
    monkeypatch.setenv("KANBUS_REALTIME_UDS_SOCKET_PATH", "/tmp/kanbus.sock")
    monkeypatch.setenv(
        "KANBUS_REALTIME_TOPICS_PROJECT_EVENTS", "projects/{project}/events"
    )
    monkeypatch.setenv("KANBUS_OVERLAY_ENABLED", "1")
    monkeypatch.setenv("KANBUS_OVERLAY_TTL_S", "42")

    merged: dict = {}
    config_loader._apply_environment_overrides(merged)
    assert merged["realtime"]["transport"] == "mqtt"
    assert merged["realtime"]["broker"] == "mqtt://localhost:1883"
    assert merged["realtime"]["autostart"] is True
    assert merged["realtime"]["keepalive"] is False
    assert merged["realtime"]["uds_socket_path"] == "/tmp/kanbus.sock"
    assert merged["realtime"]["topics"]["project_events"] == "projects/{project}/events"
    assert merged["overlay"]["enabled"] is True
    assert merged["overlay"]["ttl_s"] == 42


def test_has_unknown_fields_detects_extra_forbidden() -> None:
    try:
        RealtimeConfig.model_validate({"unknown_field": "x"})
    except ValidationError as error:
        assert config_loader._has_unknown_fields(error) is True
    else:
        raise AssertionError("expected ValidationError")



from __future__ import annotations

from pathlib import Path

import click
import pytest

from kanbus import agents_management as agents
from kanbus.config_loader import ConfigurationError
from kanbus.models import PriorityDefinition, StatusDefinition
from kanbus.project import ProjectMarkerError

from test_helpers import build_project_configuration


def _config():
    cfg = build_project_configuration().model_copy(
        update={
            "project_key": "KB",
            "project_management_template": None,
            "transition_labels": {
                "default": {
                    "open": {"in_progress": "start", "closed": "close"},
                    "in_progress": {"open": "reopen", "closed": "finish"},
                    "closed": {"open": "reopen"},
                }
            },
            "statuses": [
                StatusDefinition(key="open", name="Open", category="To do"),
                StatusDefinition(
                    key="in_progress", name="In progress", category="In progress"
                ),
                StatusDefinition(key="closed", name="Closed", category="Done"),
                StatusDefinition(key="blocked", name="Blocked", category="In progress"),
            ],
            "priorities": {
                0: PriorityDefinition(name="critical"),
                2: PriorityDefinition(name="medium"),
            },
            "default_priority": 2,
            "hierarchy": ["initiative", "epic", "task"],
            "types": ["bug", "story", "chore"],
        }
    )
    return cfg


def test_ensure_agents_file_creates_new_files(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(agents, "build_project_management_text", lambda _root: "PM")
    calls: list[str] = []
    monkeypatch.setattr(
        agents, "_write_project_guard_files", lambda _p: calls.append("guard")
    )
    monkeypatch.setattr(
        agents, "_write_tool_block_files", lambda _p: calls.append("tools")
    )

    changed = agents.ensure_agents_file(tmp_path, force=False)
    assert changed is True
    assert (tmp_path / "AGENTS.md").exists()
    assert (tmp_path / agents.PROJECT_MANAGEMENT_FILENAME).read_text(
        encoding="utf-8"
    ) == "PM"
    assert calls == []


def test_ensure_agents_file_existing_with_match_no_force_and_decline(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    agents_path = tmp_path / "AGENTS.md"
    agents_path.write_text("# Agent Instructions\n\n## Kanbus\nold\n", encoding="utf-8")

    monkeypatch.setattr(agents, "build_project_management_text", lambda _root: "PM")
    monkeypatch.setattr(agents, "_confirm_overwrite", lambda: False)

    calls: list[str] = []
    monkeypatch.setattr(
        agents, "_write_project_guard_files", lambda _p: calls.append("guard")
    )
    monkeypatch.setattr(
        agents, "_write_tool_block_files", lambda _p: calls.append("tools")
    )

    changed = agents.ensure_agents_file(tmp_path, force=False)
    assert changed is False
    assert calls == ["guard", "tools"]


def test_ensure_agents_file_existing_with_match_force_replaces(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    agents_path = tmp_path / "AGENTS.md"
    agents_path.write_text("# Agent Instructions\n\n## Kanbus\nold\n", encoding="utf-8")

    monkeypatch.setattr(agents, "build_project_management_text", lambda _root: "PM")
    monkeypatch.setattr(agents, "_write_project_guard_files", lambda _p: None)
    monkeypatch.setattr(agents, "_write_tool_block_files", lambda _p: None)

    changed = agents.ensure_agents_file(tmp_path, force=True)
    assert changed is True
    text = agents_path.read_text(encoding="utf-8")
    assert agents.KANBUS_SECTION_HEADER in text
    assert "old" not in text


def test_ensure_agents_file_existing_without_match_inserts(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    agents_path = tmp_path / "AGENTS.md"
    agents_path.write_text("# Agent Instructions\n\n## Other\ntext\n", encoding="utf-8")

    monkeypatch.setattr(agents, "build_project_management_text", lambda _root: "PM")
    monkeypatch.setattr(agents, "_write_project_guard_files", lambda _p: None)
    monkeypatch.setattr(agents, "_write_tool_block_files", lambda _p: None)

    changed = agents.ensure_agents_file(tmp_path, force=False)
    assert changed is True
    assert agents.KANBUS_SECTION_HEADER in agents_path.read_text(encoding="utf-8")


def test_ensure_project_management_file_honors_force(tmp_path: Path) -> None:
    p = tmp_path / agents.PROJECT_MANAGEMENT_FILENAME
    p.write_text("old", encoding="utf-8")

    agents._ensure_project_management_file(
        tmp_path, force=False, instructions_text="new"
    )
    assert p.read_text(encoding="utf-8") == "old"

    agents._ensure_project_management_file(
        tmp_path, force=True, instructions_text="new"
    )
    assert p.read_text(encoding="utf-8") == "new"


def test_build_project_management_text_default_custom_and_template_error(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = _config()
    monkeypatch.setattr(
        agents, "_load_configuration_for_instructions", lambda _root: cfg
    )

    rendered = agents.build_project_management_text(tmp_path)
    assert "KB" in rendered

    custom_template = tmp_path / "custom-template.md"
    custom_template.write_text("Project {{ project_key }}", encoding="utf-8")
    cfg_custom = cfg.model_copy(
        update={"project_management_template": str(custom_template)}
    )
    monkeypatch.setattr(
        agents, "_load_configuration_for_instructions", lambda _root: cfg_custom
    )
    rendered_custom = agents.build_project_management_text(tmp_path)
    assert rendered_custom == "Project KB"

    bad_template = tmp_path / "bad.md"
    bad_template.write_text("{% for x in %}", encoding="utf-8")
    cfg_bad = cfg.model_copy(update={"project_management_template": str(bad_template)})
    monkeypatch.setattr(
        agents, "_load_configuration_for_instructions", lambda _root: cfg_bad
    )
    with pytest.raises(click.ClickException):
        agents.build_project_management_text(tmp_path)


def test_load_configuration_for_instructions_wraps_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(
        agents,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ProjectMarkerError("missing")),
    )
    with pytest.raises(click.ClickException, match="missing"):
        agents._load_configuration_for_instructions(tmp_path)

    monkeypatch.setattr(
        agents, "get_configuration_path", lambda _root: tmp_path / ".kanbus.yml"
    )
    monkeypatch.setattr(
        agents,
        "load_project_configuration",
        lambda _path: (_ for _ in ()).throw(ConfigurationError("bad")),
    )
    with pytest.raises(click.ClickException, match="bad"):
        agents._load_configuration_for_instructions(tmp_path)


def test_resolve_project_management_template_path_paths(
    tmp_path: Path,
) -> None:
    cfg = _config()

    assert agents._resolve_project_management_template_path(tmp_path, cfg) is None

    conventional = tmp_path / agents.DEFAULT_PROJECT_MANAGEMENT_TEMPLATE_FILENAME
    conventional.write_text("x", encoding="utf-8")
    assert (
        agents._resolve_project_management_template_path(tmp_path, cfg) == conventional
    )

    custom = tmp_path / "custom.md"
    custom.write_text("x", encoding="utf-8")
    cfg_custom = cfg.model_copy(update={"project_management_template": "custom.md"})
    assert (
        agents._resolve_project_management_template_path(tmp_path, cfg_custom) == custom
    )

    cfg_missing = cfg.model_copy(update={"project_management_template": "missing.md"})
    with pytest.raises(click.ClickException, match="template not found"):
        agents._resolve_project_management_template_path(tmp_path, cfg_missing)


def test_context_and_example_builders() -> None:
    cfg = _config()

    workflow_context = agents._build_workflow_context(cfg)
    assert workflow_context
    assert workflow_context[0]["name"] == "default"

    priorities = agents._build_priority_context(cfg.priorities)
    assert priorities[0]["value"] == 0

    examples = agents._build_command_examples(cfg)
    assert any("kanbus create" in line for line in examples)
    assert any("kanbus close" in line for line in examples)

    semantic = agents._build_semantic_release_mapping(
        ["bug", "story", "chore", "other"]
    )
    assert semantic[0]["category"] == "fix"
    assert semantic[1]["category"] == "feat"

    statuses = agents._collect_statuses(cfg.workflows["default"])
    assert "open" in statuses
    assert agents._select_status_example("open", cfg.workflows["default"]) in statuses

    no_transition_workflow = {"x": []}
    assert agents._select_status_example("x", no_transition_workflow) == "x"

    ctx = agents._build_project_management_context(cfg)
    assert ctx["project_key"] == "KB"
    assert ctx["default_priority_name"] == "medium"

    assert agents._build_parent_child_rules(["task"], ["bug"]) == [
        "bug cannot have parents."
    ]

    cfg_blocked = cfg.model_copy(
        update={
            "workflows": {
                "default": {
                    "open": ["in_progress", "blocked"],
                    "in_progress": ["open", "closed"],
                    "closed": ["open"],
                }
            }
        }
    )
    examples_blocked = agents._build_command_examples(cfg_blocked)
    assert "kanbus update <id> --status blocked" in examples_blocked

    workflow_with_transitions = {"todo": ["done"], "done": []}
    assert agents._select_status_example("missing", workflow_with_transitions) == "done"


def test_header_section_utilities_and_replace_insert_paths() -> None:
    lines = ["# Header", "", "## A", "a", "## B", "b", "### C", "c"]
    end = agents._find_section_end(lines, 3, 2)
    assert end == 4

    matches = [
        agents.SectionMatch(start=2, end=4, level=2),
        agents.SectionMatch(start=4, end=8, level=2),
    ]
    assert agents._is_in_sections(2, matches) is True
    assert agents._is_in_sections(1, matches) is False

    replaced = agents._replace_sections(lines, matches, matches[0], ["## X", "new"])
    assert "## X" in replaced

    replaced_no_insert = agents._replace_sections(
        lines, [], agents.SectionMatch(0, 0, 1), ["## X"]
    )
    assert replaced_no_insert.endswith("## X\n")

    inserted = agents._insert_kanbus_section(["plain"], ["## K"])
    assert inserted.startswith("## K")
    assert agents._parse_header("#     ") is None


def test_confirm_overwrite_tty_and_abort_paths(monkeypatch: pytest.MonkeyPatch) -> None:
    monkeypatch.delenv("KANBUS_NON_INTERACTIVE", raising=False)
    monkeypatch.delenv("KANBUS_FORCE_INTERACTIVE", raising=False)

    class _NoTty:
        def isatty(self):
            return False

    monkeypatch.setattr(agents.click, "get_text_stream", lambda _name: _NoTty())
    with pytest.raises(click.ClickException, match="Re-run with --force"):
        agents._confirm_overwrite()

    class _Tty:
        def isatty(self):
            return True

    monkeypatch.setattr(agents.click, "get_text_stream", lambda _name: _Tty())
    monkeypatch.setattr(
        agents.click,
        "confirm",
        lambda _prompt, default: (_ for _ in ()).throw(click.Abort()),
    )
    with pytest.raises(click.ClickException, match="Re-run with --force"):
        agents._confirm_overwrite()



"""Overlay cache helpers for speculative realtime updates."""

from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timedelta, timezone
import json
from pathlib import Path
import subprocess
import stat
from typing import Iterable, Optional

from kanbus.config_loader import load_project_configuration
from kanbus.models import IssueData, OverlayConfig
from kanbus.project import get_configuration_path, resolve_labeled_projects


@dataclass(frozen=True)
class OverlayIssueRecord:
    """Overlay issue record."""

    issue: IssueData
    overlay_ts: str
    overlay_event_id: Optional[str]
    overrides: Optional[dict[str, object]] = None


@dataclass(frozen=True)
class OverlayTombstone:
    """Overlay tombstone record."""

    op: str
    project: str
    id: str
    event_id: Optional[str]
    ts: str
    ttl_s: int


def overlay_root(project_dir: Path) -> Path:
    """Return the overlay root directory for a project."""
    return project_dir / ".overlay"


def overlay_issue_path(project_dir: Path, issue_id: str) -> Path:
    return overlay_root(project_dir) / "issues" / f"{issue_id}.json"


def overlay_tombstone_path(project_dir: Path, issue_id: str) -> Path:
    return overlay_root(project_dir) / "tombstones" / f"{issue_id}.json"


def write_overlay_issue(
    project_dir: Path,
    issue: IssueData,
    overlay_ts: str,
    overlay_event_id: Optional[str],
) -> None:
    """Write an overlay issue snapshot."""
    payload = {
        "overlay_ts": overlay_ts,
        "overlay_event_id": overlay_event_id,
        "overrides": None,
        "issue": issue.model_dump(by_alias=True, mode="json"),
    }
    path = overlay_issue_path(project_dir, issue.identifier)
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(json.dumps(payload, indent=2), encoding="utf-8")


def write_tombstone(
    project_dir: Path,
    tombstone: OverlayTombstone,
) -> None:
    """Write an overlay tombstone."""
    path = overlay_tombstone_path(project_dir, tombstone.id)
    path.parent.mkdir(parents=True, exist_ok=True)
    payload = {
        "op": tombstone.op,
        "project": tombstone.project,
        "id": tombstone.id,
        "event_id": tombstone.event_id,
        "ts": tombstone.ts,
        "ttl_s": tombstone.ttl_s,
    }
    path.write_text(json.dumps(payload, indent=2), encoding="utf-8")


def load_overlay_issue(
    project_dir: Path, issue_id: str
) -> Optional[OverlayIssueRecord]:
    path = overlay_issue_path(project_dir, issue_id)
    if not path.exists():
        return None
    payload = json.loads(path.read_text(encoding="utf-8"))
    issue = IssueData.model_validate(payload.get("issue", {}))
    return OverlayIssueRecord(
        issue=issue,
        overlay_ts=payload.get("overlay_ts", ""),
        overlay_event_id=payload.get("overlay_event_id"),
        overrides=payload.get("overrides"),
    )


@dataclass(frozen=True)
class OverlayReconcileStats:
    """Aggregate overlay reconcile results."""

    projects: int = 0
    issues_scanned: int = 0
    issues_updated: int = 0
    issues_removed: int = 0
    fields_pruned: int = 0


def load_tombstone(project_dir: Path, issue_id: str) -> Optional[OverlayTombstone]:
    path = overlay_tombstone_path(project_dir, issue_id)
    if not path.exists():
        return None
    payload = json.loads(path.read_text(encoding="utf-8"))
    return OverlayTombstone(
        op=payload.get("op", "delete"),
        project=payload.get("project", ""),
        id=payload.get("id", issue_id),
        event_id=payload.get("event_id"),
        ts=payload.get("ts", ""),
        ttl_s=int(payload.get("ttl_s", 0) or 0),
    )


def resolve_issue_with_overlay(
    project_dir: Path,
    base_issue: Optional[IssueData],
    overlay_issue: Optional[OverlayIssueRecord],
    tombstone: Optional[OverlayTombstone],
    config: OverlayConfig,
    project_label: Optional[str] = None,
) -> Optional[IssueData]:
    """Resolve an issue with overlay and tombstone semantics."""
    if not config.enabled:
        return base_issue

    now = datetime.now(timezone.utc)
    base_updated = base_issue.updated_at if base_issue else None
    base_event_id = None
    if base_issue is not None:
        candidate = base_issue.custom.get("last_event_id")
        if isinstance(candidate, str):
            base_event_id = candidate

    if tombstone:
        if _is_expired(tombstone.ts, tombstone.ttl_s, now):
            _remove_path(overlay_tombstone_path(project_dir, tombstone.id))
            tombstone = None

    if tombstone and _tombstone_newer_than_base(tombstone.ts, base_updated):
        return None

    if overlay_issue:
        if _is_expired(overlay_issue.overlay_ts, config.ttl_s, now):
            _remove_path(
                overlay_issue_path(project_dir, overlay_issue.issue.identifier)
            )
        elif base_updated and _overlay_is_newer(
            overlay_issue.overlay_ts,
            base_updated,
            overlay_issue.overlay_event_id,
            base_event_id,
        ):
            if (
                base_issue is not None
                and overlay_issue.overrides
                and isinstance(overlay_issue.overrides, dict)
            ):
                merged = _apply_overrides(base_issue, overlay_issue.overrides)
                if merged is not None:
                    return _tag_issue(merged, project_label)
            return _tag_issue(overlay_issue.issue, project_label)
        elif base_updated and not _overlay_is_newer(
            overlay_issue.overlay_ts,
            base_updated,
            overlay_issue.overlay_event_id,
            base_event_id,
        ):
            _remove_path(
                overlay_issue_path(project_dir, overlay_issue.issue.identifier)
            )
        elif base_issue is None:
            return _tag_issue(overlay_issue.issue, project_label)

    return _tag_issue(base_issue, project_label) if base_issue else None


def apply_overlay_to_issues(
    project_dir: Path,
    issues: Iterable[IssueData],
    config: OverlayConfig,
    project_label: Optional[str] = None,
) -> list[IssueData]:
    """Apply overlay rules to a list of issues."""
    if not config.enabled:
        return list(issues)

    base_by_id = {issue.identifier: issue for issue in issues}
    results: list[IssueData] = []

    for issue in issues:
        if issue.custom.get("source") == "local":
            results.append(issue)
            continue
        overlay_issue = load_overlay_issue(project_dir, issue.identifier)
        tombstone = load_tombstone(project_dir, issue.identifier)
        resolved = resolve_issue_with_overlay(
            project_dir,
            issue,
            overlay_issue,
            tombstone,
            config,
            project_label=project_label,
        )
        if resolved is not None:
            results.append(resolved)

    overlay_dir = overlay_root(project_dir) / "issues"
    if overlay_dir.exists():
        for entry in overlay_dir.glob("*.json"):
            issue_id = entry.stem
            if issue_id in base_by_id:
                continue
            overlay_issue = load_overlay_issue(project_dir, issue_id)
            if overlay_issue is None:
                continue
            tombstone = load_tombstone(project_dir, issue_id)
            resolved = resolve_issue_with_overlay(
                project_dir,
                None,
                overlay_issue,
                tombstone,
                config,
                project_label=project_label,
            )
            if resolved is not None:
                results.append(resolved)

    results.sort(key=lambda issue: issue.identifier)
    return results


def gc_overlay(project_dir: Path, config: OverlayConfig) -> None:
    """Sweep overlay entries and remove stale or expired records."""
    if not config.enabled:
        return
    now = datetime.now(timezone.utc)
    issues_dir = overlay_root(project_dir) / "issues"
    tombstones_dir = overlay_root(project_dir) / "tombstones"

    if issues_dir.exists():
        for entry in issues_dir.glob("*.json"):
            issue_id = entry.stem
            overlay_issue = load_overlay_issue(project_dir, issue_id)
            if overlay_issue is None:
                _remove_path(entry)
                continue
            base_path = project_dir / "issues" / f"{issue_id}.json"
            base_issue = None
            if base_path.exists():
                try:
                    base_issue = IssueData.model_validate(
                        json.loads(base_path.read_text(encoding="utf-8"))
                    )
                except Exception:
                    base_issue = None
            if _is_expired(overlay_issue.overlay_ts, config.ttl_s, now):
                _remove_path(entry)
            elif base_issue and not _overlay_is_newer(
                overlay_issue.overlay_ts,
                base_issue.updated_at,
                overlay_issue.overlay_event_id,
                _extract_event_id(base_issue),
            ):
                _remove_path(entry)

    if tombstones_dir.exists():
        for entry in tombstones_dir.glob("*.json"):
            issue_id = entry.stem
            tombstone = load_tombstone(project_dir, issue_id)
            if tombstone is None:
                _remove_path(entry)
                continue
            base_path = project_dir / "issues" / f"{issue_id}.json"
            base_issue = None
            if base_path.exists():
                try:
                    base_issue = IssueData.model_validate(
                        json.loads(base_path.read_text(encoding="utf-8"))
                    )
                except Exception:
                    base_issue = None
            if _is_expired(tombstone.ts, tombstone.ttl_s, now):
                _remove_path(entry)
            elif base_issue and _base_newer_than_tombstone(
                base_issue.updated_at, tombstone.ts
            ):
                _remove_path(entry)


def reconcile_overlay(
    project_dir: Path,
    config: OverlayConfig,
    prune: bool,
    dry_run: bool,
) -> OverlayReconcileStats:
    """Reconcile overlay snapshots against canonical issue files."""
    stats = OverlayReconcileStats()
    if not config.enabled:
        return stats
    issues_dir = overlay_root(project_dir) / "issues"
    if not issues_dir.exists():
        return stats

    issues_scanned = 0
    issues_updated = 0
    issues_removed = 0
    fields_pruned = 0

    for entry in issues_dir.glob("*.json"):
        issue_id = entry.stem
        overlay_issue = load_overlay_issue(project_dir, issue_id)
        if overlay_issue is None:
            if not dry_run:
                _remove_path(entry)
            continue
        base_path = project_dir / "issues" / f"{issue_id}.json"
        if not base_path.exists():
            continue
        try:
            base_issue = IssueData.model_validate(
                json.loads(base_path.read_text(encoding="utf-8"))
            )
        except Exception:
            continue

        issues_scanned += 1
        overrides = (
            dict(overlay_issue.overrides)
            if isinstance(overlay_issue.overrides, dict)
            else _diff_issue_fields(base_issue, overlay_issue.issue)
        )
        if prune:
            before = len(overrides)
            base_values = _issue_to_map(base_issue)
            overrides = {
                key: value
                for key, value in overrides.items()
                if base_values.get(key) != value
            }
            fields_pruned += before - len(overrides)
        if not overrides:
            issues_removed += 1
            if not dry_run:
                _remove_path(entry)
            continue

        merged_issue = _apply_overrides(base_issue, overrides)
        if merged_issue is None:
            continue
        needs_write = merged_issue != overlay_issue.issue or dict(overrides) != dict(
            overlay_issue.overrides or {}
        )
        if needs_write:
            issues_updated += 1
            if not dry_run:
                payload = {
                    "overlay_ts": overlay_issue.overlay_ts,
                    "overlay_event_id": overlay_issue.overlay_event_id,
                    "overrides": overrides,
                    "issue": merged_issue.model_dump(by_alias=True, mode="json"),
                }
                entry.write_text(json.dumps(payload, indent=2), encoding="utf-8")

    return OverlayReconcileStats(
        projects=0,
        issues_scanned=issues_scanned,
        issues_updated=issues_updated,
        issues_removed=issues_removed,
        fields_pruned=fields_pruned,
    )


def _tag_issue(
    issue: Optional[IssueData], project_label: Optional[str]
) -> Optional[IssueData]:
    if issue is None:
        return None
    custom = dict(issue.custom)
    custom.setdefault("source", "shared")
    if project_label:
        custom.setdefault("project_label", project_label)
    return issue.model_copy(update={"custom": custom})


def _parse_ts(timestamp: str) -> Optional[datetime]:
    if not timestamp:
        return None
    try:
        return datetime.fromisoformat(timestamp.replace("Z", "+00:00"))
    except ValueError:
        return None


def _is_expired(timestamp: str, ttl_s: int, now: datetime) -> bool:
    if ttl_s <= 0:
        return False
    parsed = _parse_ts(timestamp)
    if parsed is None:
        return False
    return parsed + timedelta(seconds=ttl_s) < now


def _overlay_is_newer(
    overlay_ts: str,
    base_updated: datetime,
    overlay_event_id: Optional[str],
    base_event_id: Optional[str],
) -> bool:
    parsed = _parse_ts(overlay_ts)
    if parsed is None:
        return False
    if parsed > base_updated:
        return True
    if parsed < base_updated:
        return False
    if overlay_event_id and base_event_id:
        return overlay_event_id > base_event_id
    return True


def _tombstone_newer_than_base(
    tombstone_ts: str, base_updated: Optional[datetime]
) -> bool:
    parsed = _parse_ts(tombstone_ts)
    if parsed is None:
        return False
    if base_updated is None:
        return True
    return parsed >= base_updated


def _base_newer_than_tombstone(base_updated: datetime, tombstone_ts: str) -> bool:
    parsed = _parse_ts(tombstone_ts)
    if parsed is None:
        return True
    return base_updated > parsed


def _extract_event_id(issue: IssueData) -> Optional[str]:
    candidate = issue.custom.get("last_event_id")
    if isinstance(candidate, str):
        return candidate
    return None


def _remove_path(path: Path) -> None:
    try:
        path.unlink(missing_ok=True)
    except OSError:
        return


def gc_overlay_for_projects(
    root: Path, project_label: Optional[str], all_projects: bool
) -> int:
    """Run overlay GC for one or more projects."""
    if project_label and all_projects:
        raise ValueError("cannot combine --project with --all")
    configuration = load_project_configuration(get_configuration_path(root))
    labeled = resolve_labeled_projects(root)
    if not labeled:
        return 0
    if all_projects:
        selected = labeled
    else:
        label = project_label or configuration.project_key
        selected = [project for project in labeled if project.label == label]
        if not selected:
            raise ValueError(f"unknown project label: {label}")
    for project in selected:
        gc_overlay(project.project_dir, configuration.overlay)
    return len(selected)


def reconcile_overlay_for_projects(
    root: Path,
    project_label: Optional[str],
    all_projects: bool,
    prune: bool,
    dry_run: bool,
) -> OverlayReconcileStats:
    """Run overlay reconcile for one or more projects."""
    if project_label and all_projects:
        raise ValueError("cannot combine --project with --all")
    configuration = load_project_configuration(get_configuration_path(root))
    labeled = resolve_labeled_projects(root)
    if not labeled:
        return OverlayReconcileStats()
    if all_projects:
        selected = labeled
    else:
        label = project_label or configuration.project_key
        selected = [project for project in labeled if project.label == label]
        if not selected:
            raise ValueError(f"unknown project label: {label}")
    aggregate = OverlayReconcileStats(projects=len(selected))
    for project in selected:
        stats = reconcile_overlay(
            project.project_dir, configuration.overlay, prune=prune, dry_run=dry_run
        )
        aggregate = OverlayReconcileStats(
            projects=aggregate.projects,
            issues_scanned=aggregate.issues_scanned + stats.issues_scanned,
            issues_updated=aggregate.issues_updated + stats.issues_updated,
            issues_removed=aggregate.issues_removed + stats.issues_removed,
            fields_pruned=aggregate.fields_pruned + stats.fields_pruned,
        )
    return aggregate


def install_overlay_hooks(root: Path) -> None:
    """Install git hooks to run overlay reconcile/gc after git operations."""
    hooks_dir = _resolve_git_hooks_dir(root)
    hooks_dir.mkdir(parents=True, exist_ok=True)
    hook_block = _overlay_hook_block()
    for hook in ("post-merge", "post-checkout", "post-rewrite"):
        hook_path = hooks_dir / hook
        if hook_path.exists():
            existing = hook_path.read_text(encoding="utf-8")
            if (
                "Kanbus overlay cache maintenance" in existing
                or "Kanbus overlay cache GC" in existing
            ):
                continue
            contents = existing.rstrip() + "\n\n" + hook_block
        else:
            contents = "#!/bin/sh\n" + hook_block
        hook_path.write_text(contents + "\n", encoding="utf-8")
        _ensure_executable(hook_path)


def _resolve_git_hooks_dir(root: Path) -> Path:
    result = subprocess.run(
        ["git", "rev-parse", "--git-path", "hooks"],
        cwd=root,
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0:
        raise RuntimeError("not a git repository")
    path = Path(result.stdout.strip())
    if not path.is_absolute():
        path = root / path
    if path == Path("/dev/null"):
        raise RuntimeError(
            "git hooks are disabled (core.hooksPath=/dev/null); run `git config --unset core.hooksPath` to enable hook installation"
        )
    if path.exists() and not path.is_dir():
        raise RuntimeError(f"git hooks path is not a directory: {path}")
    return path


def _overlay_hook_block() -> str:
    return "\n".join(
        [
            "# Kanbus overlay cache maintenance",
            "if command -v kanbus >/dev/null 2>&1; then",
            "  kanbus overlay reconcile --all --prune >/dev/null 2>&1 || true",
            "  kanbus overlay gc --all >/dev/null 2>&1 || true",
            "elif command -v kbs >/dev/null 2>&1; then",
            "  kbs overlay reconcile --all --prune >/dev/null 2>&1 || true",
            "  kbs overlay gc --all >/dev/null 2>&1 || true",
            "fi",
        ]
    )


def _issue_to_map(issue: IssueData) -> dict[str, object]:
    return dict(issue.model_dump(by_alias=True, mode="json"))


def _issue_from_map(payload: dict[str, object]) -> Optional[IssueData]:
    try:
        return IssueData.model_validate(payload)
    except Exception:
        return None


def _apply_overrides(
    base_issue: IssueData, overrides: dict[str, object]
) -> Optional[IssueData]:
    merged = _issue_to_map(base_issue)
    merged.update(overrides)
    return _issue_from_map(merged)


def _diff_issue_fields(
    base_issue: IssueData, overlay_issue: IssueData
) -> dict[str, object]:
    base_payload = _issue_to_map(base_issue)
    overlay_payload = _issue_to_map(overlay_issue)
    return {
        key: value
        for key, value in overlay_payload.items()
        if base_payload.get(key) != value
    }


def _ensure_executable(path: Path) -> None:
    try:
        mode = path.stat().st_mode
        path.chmod(mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)
    except OSError:
        return



from __future__ import annotations

import sys
from pathlib import Path

REPO_ROOT = Path(__file__).resolve().parents[2]
sys.path.insert(0, str(REPO_ROOT / "tools"))

import check_spec_parity  # noqa: E402


def test_decode_escaped_literal_unescapes_regex_backslashes() -> None:
    literal = r'a kanbus issue (?P<identifier>[^"\\s]+) exists with priority (?P<priority>\\d+)'

    decoded = check_spec_parity._decode_escaped_literal(literal)

    assert r"\d+" in decoded
    assert r"\\d+" not in decoded


def test_collect_feature_steps_ignores_docstring_content(tmp_path: Path) -> None:
    feature_dir = tmp_path / "features"
    feature_dir.mkdir(parents=True)
    feature_path = feature_dir / "sample.feature"
    feature_path.write_text(
        (
            "Feature: Sample\n"
            "  Scenario: Ignores embedded policy text\n"
            "    Given a real step\n"
            '    """\n'
            "    Then this line should not be parsed as a step\n"
            '    """\n'
            "    Then another real step\n"
        ),
        encoding="utf-8",
    )

    steps = check_spec_parity.collect_feature_steps(feature_dir)

    assert "a real step" in steps
    assert "another real step" in steps
    assert "this line should not be parsed as a step" not in steps


def test_report_respects_non_strict_mode(monkeypatch) -> None:
    results = check_spec_parity.ParityResults(
        feature_steps={"a missing step"},
        python_steps=set(),
        rust_steps=set(),
        python_patterns=[],
        rust_patterns=[],
    )

    monkeypatch.setenv("KANBUS_PARITY_STRICT", "0")
    ok, _lines = check_spec_parity.report(results)

    assert ok is True


def test_report_fails_in_strict_mode_when_missing_steps(monkeypatch) -> None:
    results = check_spec_parity.ParityResults(
        feature_steps={"a missing step"},
        python_steps=set(),
        rust_steps=set(),
        python_patterns=[],
        rust_patterns=[],
    )

    monkeypatch.setenv("KANBUS_PARITY_STRICT", "1")
    ok, lines = check_spec_parity.report(results)

    assert ok is False
    assert any("Missing in Python:" in line for line in lines)
    assert any("Missing in Rust:" in line for line in lines)



"""
File system helpers for initialization.
"""

from __future__ import annotations

import subprocess
import json
from dataclasses import dataclass
from pathlib import Path
import yaml

from kanbus.config import DEFAULT_CONFIGURATION
from kanbus.project_management_template import (
    DEFAULT_PROJECT_MANAGEMENT_TEMPLATE,
    DEFAULT_PROJECT_MANAGEMENT_TEMPLATE_FILENAME,
)
from kanbus.wiki_templates import (
    DEFAULT_WIKI_INDEX,
    DEFAULT_WIKI_INDEX_FILENAME,
    DEFAULT_WIKI_WHATS_NEXT,
    DEFAULT_WIKI_WHATS_NEXT_FILENAME,
)
from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.project import (
    ProjectMarkerError,
    ensure_project_local_directory,
    get_configuration_path,
)


class InitializationError(RuntimeError):
    """Raised when project initialization fails."""


@dataclass
class RepairPlan:
    project_dir: Path
    missing_project_dir: bool
    missing_issues_dir: bool
    missing_events_dir: bool


def ensure_git_repository(root: Path) -> None:
    """Ensure the provided path is inside a git repository.

    :param root: Directory to check for a git repository.
    :type root: Path
    :raises InitializationError: If the path is not a git repository.
    """
    result = subprocess.run(
        ["git", "rev-parse", "--is-inside-work-tree"],
        cwd=root,
        capture_output=True,
        text=True,
        check=False,
    )
    if result.returncode != 0 or result.stdout.strip() != "true":
        raise InitializationError("not a git repository")


def initialize_project(root: Path, create_local: bool = False) -> None:
    """Initialize the Kanbus project directory structure.

    :param root: Repository root path.
    :type root: Path
    :param create_local: Whether to create a project-local directory.
    :type create_local: bool
    :raises InitializationError: If the project is already initialized.
    """
    project_dir = root / "project"
    if project_dir.exists():
        raise InitializationError("already initialized")

    issues_dir = project_dir / "issues"
    events_dir = project_dir / "events"

    project_dir.mkdir(parents=True, exist_ok=False)
    issues_dir.mkdir(parents=True)
    events_dir.mkdir(parents=True)
    config_path = root / ".kanbus.yml"
    if not config_path.exists():
        config_path.write_text(
            yaml.safe_dump(DEFAULT_CONFIGURATION, sort_keys=False),
            encoding="utf-8",
        )
    _write_project_guard_files_if_missing(project_dir)
    _write_tool_block_files(root)
    _ensure_gitignore_entry(root, "project/.overlay/")
    template_path = root / DEFAULT_PROJECT_MANAGEMENT_TEMPLATE_FILENAME
    if not template_path.exists():
        template_path.write_text(
            DEFAULT_PROJECT_MANAGEMENT_TEMPLATE,
            encoding="utf-8",
        )
    wiki_dir = project_dir / "wiki"
    wiki_dir.mkdir(parents=True)
    (wiki_dir / DEFAULT_WIKI_INDEX_FILENAME).write_text(
        DEFAULT_WIKI_INDEX, encoding="utf-8"
    )
    (wiki_dir / DEFAULT_WIKI_WHATS_NEXT_FILENAME).write_text(
        DEFAULT_WIKI_WHATS_NEXT, encoding="utf-8"
    )
    if create_local:
        ensure_project_local_directory(project_dir)


def _ensure_gitignore_entry(root: Path, entry: str) -> None:
    gitignore_path = root / ".gitignore"
    existing = ""
    if gitignore_path.exists():
        existing = gitignore_path.read_text(encoding="utf-8")
    lines = [line.strip() for line in existing.splitlines() if line.strip()]
    if entry in lines:
        return
    lines.append(entry)
    gitignore_path.write_text("\n".join(lines) + "\n", encoding="utf-8")


def _write_guard_files_in_subdir(subdir: Path, folder_name: str) -> None:
    """Write AGENTS.md and DO_NOT_EDIT into a guarded subdir (e.g. issues/ or events/)."""
    agents_path = subdir / "AGENTS.md"
    agents_path.write_text(
        "\n".join(
            [
                "# DO NOT EDIT HERE",
                "",
                f"Do not read or write in this folder ({folder_name}/). Use Kanbus commands instead.",
                "Do not inspect issue JSON with tools like cat or jq.",
                "",
                "See ../../AGENTS.md and ../../CONTRIBUTING_AGENT.md for required process.",
            ]
        )
        + "\n",
        encoding="utf-8",
    )
    do_not_edit = subdir / "DO_NOT_EDIT"
    do_not_edit.write_text(
        "\n".join(
            [
                f"DO NOT EDIT THIS FOLDER ({folder_name}/)",
                "This folder is guarded by The Way.",
                "All changes must go through Kanbus (see ../../AGENTS.md and ../../CONTRIBUTING_AGENT.md).",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_project_guard_files(project_dir: Path) -> None:
    issues_dir = project_dir / "issues"
    events_dir = project_dir / "events"
    if issues_dir.exists():
        _write_guard_files_in_subdir(issues_dir, "issues")
    if events_dir.exists():
        _write_guard_files_in_subdir(events_dir, "events")
    root_agents_path = project_dir / "AGENTS.md"
    root_agents_path.write_text(
        "\n".join(
            [
                "# Project directory",
                "",
                "Do not edit issues/ or events/ directly; use Kanbus for issues and events.",
                "You may edit wiki/ (e.g. Markdown) directly.",
                "",
                "See ../AGENTS.md and ../CONTRIBUTING_AGENT.md for required process.",
            ]
        )
        + "\n",
        encoding="utf-8",
    )


def _write_project_guard_files_if_missing(project_dir: Path) -> None:
    issues_dir = project_dir / "issues"
    events_dir = project_dir / "events"
    for subdir, folder_name in [(issues_dir, "issues"), (events_dir, "events")]:
        if not subdir.exists():
            continue
        agents_path = subdir / "AGENTS.md"
        do_not_edit = subdir / "DO_NOT_EDIT"
        if not agents_path.exists() or not do_not_edit.exists():
            _write_guard_files_in_subdir(subdir, folder_name)
    root_agents_path = project_dir / "AGENTS.md"
    if not root_agents_path.exists():
        root_agents_path.write_text(
            "\n".join(
                [
                    "# Project directory",
                    "",
                    "Do not edit issues/ or events/ directly; use Kanbus for issues and events.",
                    "You may edit wiki/ (e.g. Markdown) directly.",
                    "",
                    "See ../AGENTS.md and ../CONTRIBUTING_AGENT.md for required process.",
                ]
            )
            + "\n",
            encoding="utf-8",
        )


def _write_tool_block_files(root: Path) -> None:
    cursorignore = root / ".cursorignore"
    if not cursorignore.exists():
        cursorignore.write_text("project/issues/\nproject/events/\n", encoding="utf-8")

    claude_dir = root / ".claude"
    claude_dir.mkdir(parents=True, exist_ok=True)
    claude_settings = claude_dir / "settings.json"
    if not claude_settings.exists():
        claude_settings.write_text(
            json.dumps(
                {
                    "permissions": {
                        "deny": [
                            "Read(./project/issues/**)",
                            "Edit(./project/issues/**)",
                            "Read(./project/events/**)",
                            "Edit(./project/events/**)",
                        ]
                    }
                },
                indent=2,
            )
            + "\n",
            encoding="utf-8",
        )

    vscode_dir = root / ".vscode"
    vscode_dir.mkdir(parents=True, exist_ok=True)
    vscode_settings = vscode_dir / "settings.json"
    if not vscode_settings.exists():
        vscode_settings.write_text(
            json.dumps(
                {
                    "files.exclude": {
                        "**/project/issues/**": True,
                        "**/project/events/**": True,
                    },
                    "files.watcherExclude": {
                        "**/project/issues/**": True,
                        "**/project/events/**": True,
                    },
                    "search.exclude": {
                        "**/project/issues/**": True,
                        "**/project/events/**": True,
                    },
                },
                indent=2,
            )
            + "\n",
            encoding="utf-8",
        )


def detect_repairable_project_issues(
    root: Path, *, allow_uninitialized: bool
) -> RepairPlan | None:
    try:
        config_path = get_configuration_path(root)
    except ProjectMarkerError as error:
        if allow_uninitialized:
            return None
        raise error
    except ConfigurationError as error:
        raise error

    configuration = load_project_configuration(config_path)
    project_dir = config_path.parent / configuration.project_directory
    try:
        project_dir_stat = project_dir.stat()
    except FileNotFoundError:
        project_dir_stat = None
    missing_project_dir = project_dir_stat is None

    missing_issues_dir = False
    missing_events_dir = False
    if not missing_project_dir:
        try:
            (project_dir / "issues").stat()
        except FileNotFoundError:
            missing_issues_dir = True
        try:
            (project_dir / "events").stat()
        except FileNotFoundError:
            missing_events_dir = True

    if missing_project_dir or missing_issues_dir or missing_events_dir:
        return RepairPlan(
            project_dir=project_dir,
            missing_project_dir=missing_project_dir,
            missing_issues_dir=missing_issues_dir,
            missing_events_dir=missing_events_dir,
        )
    return None


def repair_project_structure(root: Path, plan: RepairPlan) -> None:
    if plan.missing_project_dir:
        plan.project_dir.mkdir(parents=True, exist_ok=True)
    if plan.missing_issues_dir:
        (plan.project_dir / "issues").mkdir(parents=True, exist_ok=True)
    if plan.missing_events_dir:
        (plan.project_dir / "events").mkdir(parents=True, exist_ok=True)
    if plan.project_dir.exists():
        _write_project_guard_files_if_missing(plan.project_dir)

    vscode_dir = root / ".vscode"
    vscode_dir.mkdir(parents=True, exist_ok=True)
    vscode_settings = vscode_dir / "settings.json"
    if not vscode_settings.exists():
        vscode_settings.write_text(
            json.dumps(
                {
                    "files.exclude": {
                        "**/project/issues/**": True,
                        "**/project/events/**": True,
                    },
                    "files.watcherExclude": {
                        "**/project/issues/**": True,
                        "**/project/events/**": True,
                    },
                    "search.exclude": {
                        "**/project/issues/**": True,
                        "**/project/events/**": True,
                    },
                },
                indent=2,
            )
            + "\n",
            encoding="utf-8",
        )



from __future__ import annotations

from datetime import datetime
from pathlib import Path
from types import SimpleNamespace

import pytest

from kanbus import jira_sync
from kanbus.jira_sync import JiraSyncError
from kanbus.models import JiraConfiguration

from test_helpers import build_issue


def _jira_config() -> JiraConfiguration:
    return JiraConfiguration.model_validate(
        {
            "url": "https://jira.example.com",
            "project_key": "KAN",
            "sync_direction": "pull",
            "type_mappings": {"Story": "story"},
            "field_mappings": {},
        }
    )


def _jira_issue(key: str, *, summary: str = "S", parent: str | None = None):
    fields = {
        "summary": summary,
        "description": {
            "type": "doc",
            "content": [
                {"type": "text", "text": "Hello"},
                {"type": "hardBreak"},
                {"type": "text", "text": "World"},
            ],
        },
        "issuetype": {"name": "Story"},
        "status": {"name": "In Progress"},
        "priority": {"name": "High"},
        "assignee": {"displayName": "Alice"},
        "reporter": {"displayName": "Bob"},
        "labels": ["l1"],
        "comment": {
            "comments": [
                {
                    "id": "10",
                    "author": {"displayName": "Carol"},
                    "body": "plain",
                    "created": "2026-03-09T00:00:00Z",
                }
            ]
        },
        "created": "2026-03-08T00:00:00Z",
        "updated": "2026-03-09T00:00:00Z",
        "resolutiondate": "2026-03-10T00:00:00Z",
    }
    if parent:
        fields["parent"] = {"key": parent}
    return {"key": key, "fields": fields}


def test_pull_from_jira_validates_env_and_project_structure(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = _jira_config()

    monkeypatch.delenv("JIRA_API_TOKEN", raising=False)
    monkeypatch.delenv("JIRA_USER_EMAIL", raising=False)

    with pytest.raises(JiraSyncError, match="JIRA_API_TOKEN"):
        jira_sync.pull_from_jira(tmp_path, cfg, "kanbus")

    monkeypatch.setenv("JIRA_API_TOKEN", "t")
    with pytest.raises(JiraSyncError, match="JIRA_USER_EMAIL"):
        jira_sync.pull_from_jira(tmp_path, cfg, "kanbus")

    monkeypatch.setenv("JIRA_USER_EMAIL", "u@example.com")
    monkeypatch.setattr(
        "kanbus.project.load_project_directory",
        lambda _root: tmp_path / "project",
    )
    with pytest.raises(JiraSyncError, match="issues directory does not exist"):
        jira_sync.pull_from_jira(tmp_path, cfg, "kanbus")


def test_pull_from_jira_updates_and_pulls_with_dry_run_and_write_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = _jira_config()
    monkeypatch.setenv("JIRA_API_TOKEN", "t")
    monkeypatch.setenv("JIRA_USER_EMAIL", "u@example.com")

    project_dir = tmp_path / "project"
    issues_dir = project_dir / "issues"
    issues_dir.mkdir(parents=True)

    monkeypatch.setattr(
        "kanbus.project.load_project_directory", lambda _root: project_dir
    )

    issues = [
        _jira_issue("KAN-1", summary="Existing"),
        _jira_issue("KAN-2", summary="New", parent="KAN-1"),
    ]
    monkeypatch.setattr(jira_sync, "_fetch_all_jira_issues", lambda *_args: issues)

    monkeypatch.setattr(
        jira_sync, "list_issue_identifiers", lambda _dir: {"kanbus-aaaa111"}
    )
    monkeypatch.setattr(
        jira_sync,
        "_build_jira_key_index",
        lambda _ids, _dir: {"KAN-1": "kanbus-aaaa111"},
    )

    monkeypatch.setattr(
        jira_sync,
        "generate_issue_identifier",
        lambda _req: SimpleNamespace(identifier="kanbus-bbbb222"),
    )
    (issues_dir / "kanbus-aaaa111.json").write_text("{}", encoding="utf-8")

    existing_issue = build_issue("kanbus-aaaa111")
    read_calls: list[Path] = []
    monkeypatch.setattr(
        jira_sync,
        "read_issue_from_file",
        lambda p: read_calls.append(p) or existing_issue,
    )

    writes: list[Path] = []
    monkeypatch.setattr(
        jira_sync, "write_issue_to_file", lambda issue, path: writes.append(path)
    )

    result_dry = jira_sync.pull_from_jira(tmp_path, cfg, "kanbus", dry_run=True)
    assert result_dry.updated == 1
    assert result_dry.pulled == 1
    assert writes == []

    result = jira_sync.pull_from_jira(tmp_path, cfg, "kanbus", dry_run=False)
    assert result.updated == 1
    assert result.pulled == 1
    assert len(writes) == 2
    assert any(path.name == "kanbus-aaaa111.json" for path in writes)
    assert any(path.name == "kanbus-bbbb222.json" for path in writes)
    assert read_calls

    # Existing read failure should be ignored.
    monkeypatch.setattr(
        jira_sync,
        "read_issue_from_file",
        lambda _p: (_ for _ in ()).throw(RuntimeError("boom")),
    )
    jira_sync.pull_from_jira(tmp_path, cfg, "kanbus", dry_run=True)


def test_fetch_all_jira_issues_pagination_and_error(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cfg = _jira_config()

    class Response:
        def __init__(
            self, ok: bool, payload: dict, status_code: int = 200, text: str = ""
        ):
            self.ok = ok
            self._payload = payload
            self.status_code = status_code
            self.text = text

        def json(self):
            return self._payload

    calls: list[str] = []

    def fake_get(url, auth, headers, timeout):
        calls.append(url)
        if "startAt=0" in url:
            return Response(True, {"issues": [{"key": "KAN-1"}], "total": 2})
        return Response(True, {"issues": [{"key": "KAN-2"}], "total": 2})

    monkeypatch.setattr(jira_sync.requests, "get", fake_get)
    issues = jira_sync._fetch_all_jira_issues(cfg, "u@example.com", "tok")
    assert [issue["key"] for issue in issues] == ["KAN-1", "KAN-2"]
    assert len(calls) == 2

    monkeypatch.setattr(
        jira_sync.requests,
        "get",
        lambda *args, **kwargs: Response(
            False, {}, status_code=401, text="unauthorized"
        ),
    )
    with pytest.raises(JiraSyncError, match="Jira API returned 401"):
        jira_sync._fetch_all_jira_issues(cfg, "u@example.com", "tok")


def test_build_jira_key_index_handles_read_failures(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    existing_ids = {"kanbus-1", "kanbus-2"}

    def fake_read(path: Path):
        if path.name.startswith("kanbus-1"):
            issue = build_issue("kanbus-1")
            issue.custom = {"jira_key": "KAN-1"}
            return issue
        raise RuntimeError("bad")

    monkeypatch.setattr(jira_sync, "read_issue_from_file", fake_read)
    index = jira_sync._build_jira_key_index(existing_ids, tmp_path)
    assert index == {"KAN-1": "kanbus-1"}


def test_jira_key_and_summary_extractors() -> None:
    assert jira_sync._jira_issue_key({"key": "KAN-1"}) == "KAN-1"
    assert jira_sync._jira_issue_key({}) == ""
    assert jira_sync._jira_issue_summary({"fields": {"summary": "S"}}) == "S"
    assert jira_sync._jira_issue_summary({}) == "Untitled"


def test_map_jira_to_kanbus_and_support_helpers() -> None:
    cfg = _jira_config()
    issue = _jira_issue("KAN-1", summary="Story A")
    mapped = jira_sync._map_jira_to_kanbus(issue, cfg, {"KAN-1": "kanbus-1"})

    assert mapped.title == "Story A"
    assert mapped.issue_type == "story"
    assert mapped.status == "in_progress"
    assert mapped.priority == 1
    assert mapped.assignee == "Alice"
    assert mapped.creator == "Bob"
    assert mapped.labels == ["l1"]
    assert mapped.custom["jira_key"] == "KAN-1"
    assert mapped.comments[0].author == "Carol"

    assert jira_sync._extract_adf_text(None) == ""
    assert jira_sync._extract_adf_text("plain") == "plain"
    assert "Hello" in jira_sync._extract_adf_text(issue["fields"]["description"])
    assert jira_sync._extract_adf_text(123) == ""

    assert (
        jira_sync._extract_adf_content({"content": [{"type": "text", "text": "a"}]})
        == "a"
    )
    assert (
        jira_sync._extract_adf_content(
            {
                "content": [
                    {"type": "paragraph", "content": [{"type": "text", "text": "b"}]}
                ]
            }
        )
        == "b"
    )

    comments = jira_sync._extract_comments(
        {"comments": [{"author": {}, "body": None, "created": "bad-date"}]}
    )
    assert comments[0].author == "Unknown"
    assert comments[0].text == "(empty)"

    assert jira_sync._map_jira_status("To Do") == "open"
    assert jira_sync._map_jira_status("In Development") == "in_progress"
    assert jira_sync._map_jira_status("Resolved") == "closed"
    assert jira_sync._map_jira_status("Blocked") == "blocked"
    assert jira_sync._map_jira_status("SomethingElse") == "open"

    assert jira_sync._map_jira_priority("Highest") == 0
    assert jira_sync._map_jira_priority("High") == 1
    assert jira_sync._map_jira_priority("Normal") == 2
    assert jira_sync._map_jira_priority("Low") == 3
    assert jira_sync._map_jira_priority("Trivial") == 4
    assert jira_sync._map_jira_priority("Unknown") == 2

    dt = jira_sync._parse_jira_datetime("2026-03-09T00:00:00Z")
    assert isinstance(dt, datetime)
    assert dt.tzinfo is not None
    assert jira_sync._parse_jira_datetime("") is None
    assert jira_sync._parse_jira_datetime(None) is None
    assert jira_sync._parse_jira_datetime("bad") is None

    assert jira_sync.issue_path_for_identifier(Path("/x"), "kanbus-1") == Path(
        "/x/kanbus-1.json"
    )



from __future__ import annotations

import json
import subprocess
from pathlib import Path

import pytest

from kanbus import hooks
from kanbus.hooks import (
    HookEvent,
    HookExecutionError,
    HookPhase,
    HookResult,
)
from kanbus.models import HookDefinition, HooksConfiguration
from kanbus.project import ProjectMarkerError

from test_helpers import build_issue, build_project_configuration


def _hook(
    hook_id: str,
    command: list[str],
    *,
    blocking: bool | None = None,
    timeout_ms: int | None = None,
    cwd: str | None = None,
    env: dict[str, str] | None = None,
) -> HookDefinition:
    payload: dict[str, object] = {"id": hook_id, "command": command}
    if blocking is not None:
        payload["blocking"] = blocking
    if timeout_ms is not None:
        payload["timeout_ms"] = timeout_ms
    if cwd is not None:
        payload["cwd"] = cwd
    if env is not None:
        payload["env"] = env
    return HookDefinition.model_validate(payload)


def _hooks_config(
    before: dict[str, list[HookDefinition]] | None = None,
    after: dict[str, list[HookDefinition]] | None = None,
    *,
    enabled: bool = True,
    run_in_beads_mode: bool = True,
    default_timeout_ms: int = 5000,
) -> HooksConfiguration:
    return HooksConfiguration.model_validate(
        {
            "enabled": enabled,
            "run_in_beads_mode": run_in_beads_mode,
            "default_timeout_ms": default_timeout_ms,
            "before": before or {},
            "after": after or {},
        }
    )


def test_hook_invocation_to_payload_and_serialize_issue() -> None:
    invocation = hooks.HookInvocation(
        schema_version=hooks.HOOK_SCHEMA_VERSION,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_CREATE,
        timestamp="2026-03-09T00:00:00.000Z",
        actor="dev",
        mode={"runtime": "python"},
        operation={"issue_id": "kanbus-1"},
    )
    payload = invocation.to_payload()
    assert payload["phase"] == "before"
    assert payload["event"] == "issue.create"

    issue = build_issue("kanbus-1")
    issue_payload = hooks.serialize_issue(issue)
    assert issue_payload["id"] == "kanbus-1"
    assert hooks.serialize_issue(None) is None


def test_hook_handler_protocol_method_is_noop_callable() -> None:
    invocation = hooks.HookInvocation(
        schema_version=hooks.HOOK_SCHEMA_VERSION,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_CREATE,
        timestamp="2026-03-09T00:00:00.000Z",
        actor="dev",
        mode={},
        operation={},
    )
    hook = _hook("noop", ["echo"])
    assert (
        hooks.HookHandler.run(  # type: ignore[misc]
            object(),
            hook=hook,
            invocation=invocation,
            project_root=Path("."),
            timeout_ms=1,
        )
        is None
    )


def test_hooks_globally_disabled_flag_and_env(monkeypatch: pytest.MonkeyPatch) -> None:
    assert hooks.hooks_globally_disabled(no_hooks=True) is True

    monkeypatch.setenv("KANBUS_NO_HOOKS", "1")
    assert hooks.hooks_globally_disabled() is True
    monkeypatch.setenv("KANBUS_NO_HOOKS", "YES")
    assert hooks.hooks_globally_disabled() is True
    monkeypatch.setenv("KANBUS_NO_HOOKS", "off")
    assert hooks.hooks_globally_disabled() is False


def test_external_command_hook_handler_success_failure_timeout_and_oserror(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    handler = hooks.ExternalCommandHookHandler()
    invocation = hooks.HookInvocation(
        schema_version=hooks.HOOK_SCHEMA_VERSION,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_UPDATE,
        timestamp="2026-03-09T00:00:00.000Z",
        actor="dev",
        mode={"runtime": "python"},
        operation={"x": 1},
    )
    hook = _hook("h1", ["cmd"], cwd=".", env={"X": "1"})

    calls: list[dict[str, object]] = []

    def ok_run(*_args, **kwargs):
        calls.append(kwargs)
        return subprocess.CompletedProcess(
            args=["cmd"], returncode=0, stdout="", stderr=""
        )

    monkeypatch.setattr(hooks.subprocess, "run", ok_run)
    result = handler.run(
        hook=hook, invocation=invocation, project_root=tmp_path, timeout_ms=200
    )
    assert result.succeeded is True
    assert result.timed_out is False
    assert result.message == "ok"
    assert calls[0]["cwd"] == tmp_path
    assert calls[0]["env"]["X"] == "1"
    assert json.loads(calls[0]["input"])["event"] == "issue.update"

    def fail_run(*_args, **_kwargs):
        return subprocess.CompletedProcess(
            args=["cmd"], returncode=7, stdout="", stderr="bad"
        )

    monkeypatch.setattr(hooks.subprocess, "run", fail_run)
    fail = handler.run(
        hook=_hook("h2", ["cmd"]),
        invocation=invocation,
        project_root=tmp_path,
        timeout_ms=10,
    )
    assert fail.succeeded is False
    assert fail.exit_code == 7
    assert fail.message == "bad"

    def fail_no_stderr(*_args, **_kwargs):
        return subprocess.CompletedProcess(
            args=["cmd"], returncode=8, stdout="", stderr=""
        )

    monkeypatch.setattr(hooks.subprocess, "run", fail_no_stderr)
    fail2 = handler.run(
        hook=_hook("h3", ["cmd"]),
        invocation=invocation,
        project_root=tmp_path,
        timeout_ms=10,
    )
    assert fail2.message == "exit code 8"

    def timeout_run(*_args, **_kwargs):
        raise subprocess.TimeoutExpired(cmd=["cmd"], timeout=1)

    monkeypatch.setattr(hooks.subprocess, "run", timeout_run)
    timeout = handler.run(
        hook=_hook("h4", ["cmd"]),
        invocation=invocation,
        project_root=tmp_path,
        timeout_ms=10,
    )
    assert timeout.succeeded is False
    assert timeout.timed_out is True

    def os_error_run(*_args, **_kwargs):
        raise OSError("nope")

    monkeypatch.setattr(hooks.subprocess, "run", os_error_run)
    os_error = handler.run(
        hook=_hook("h5", ["cmd"]),
        invocation=invocation,
        project_root=tmp_path,
        timeout_ms=10,
    )
    assert os_error.succeeded is False
    assert os_error.timed_out is False
    assert "nope" in os_error.message


def test_resolve_hook_configuration_success_and_fallback(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = build_project_configuration()
    config_path = tmp_path / ".kanbus.yml"

    monkeypatch.setattr(hooks, "get_configuration_path", lambda _root: config_path)
    monkeypatch.setattr(hooks, "load_project_configuration", lambda _path: cfg)

    resolved, root = hooks._resolve_hook_configuration(tmp_path)
    assert isinstance(resolved, HooksConfiguration)
    assert root == tmp_path

    monkeypatch.setattr(
        hooks,
        "get_configuration_path",
        lambda _root: (_ for _ in ()).throw(ProjectMarkerError("missing")),
    )
    fallback, fallback_root = hooks._resolve_hook_configuration(tmp_path)
    assert isinstance(fallback, HooksConfiguration)
    assert fallback_root == tmp_path


def test_hooks_for_event_and_blocking_helpers() -> None:
    before_hook = _hook("b", ["echo"])
    after_hook = _hook("a", ["echo"], blocking=True)
    cfg = _hooks_config(
        before={HookEvent.ISSUE_CREATE.value: [before_hook]},
        after={HookEvent.ISSUE_CREATE.value: [after_hook]},
    )

    assert hooks._hooks_for_event(cfg, HookPhase.BEFORE, HookEvent.ISSUE_CREATE) == [
        before_hook
    ]
    assert hooks._hooks_for_event(cfg, HookPhase.AFTER, HookEvent.ISSUE_CREATE) == [
        after_hook
    ]

    assert (
        hooks._effective_blocking(HookPhase.AFTER, HookEvent.ISSUE_CREATE, before_hook)
        is False
    )
    assert (
        hooks._effective_blocking(
            HookPhase.BEFORE,
            HookEvent.ISSUE_CREATE,
            _hook("x", ["echo"], blocking=True),
        )
        is True
    )
    assert (
        hooks._effective_blocking(
            HookPhase.BEFORE, HookEvent.ISSUE_SHOW, _hook("x2", ["echo"])
        )
        is False
    )
    assert (
        hooks._effective_blocking(
            HookPhase.BEFORE, HookEvent.ISSUE_CREATE, _hook("x3", ["echo"])
        )
        is True
    )

    events = hooks._policy_events()
    assert events == sorted(events, key=lambda e: e.value)


def test_list_hooks_includes_external_and_builtin_rows(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    cfg = _hooks_config(
        before={
            "issue.create": [_hook("b1", ["echo", "ok"])],
            "unknown": [_hook("b2", ["echo", "ok"])],
        },
        after={"issue.update": [_hook("a1", ["echo", "ok"], timeout_ms=10)]},
        default_timeout_ms=55,
    )
    monkeypatch.setattr(
        hooks, "_resolve_hook_configuration", lambda _root: (cfg, Path("/repo"))
    )

    rows = hooks.list_hooks(Path("/repo"))
    assert any(
        row["source"] == "external" and row["id"] == "b1" and row["timeout_ms"] == 55
        for row in rows
    )
    assert any(
        row["source"] == "external" and row["id"] == "a1" and row["timeout_ms"] == 10
        for row in rows
    )
    assert any(
        row["source"] == "built-in" and row["id"] == "policy-guidance" for row in rows
    )


def test_validate_hooks_reports_all_validation_problems(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    script = tmp_path / "scripts" / "ok.sh"
    script.parent.mkdir(parents=True)
    script.write_text("echo ok", encoding="utf-8")

    cfg = _hooks_config(
        before={
            "bad.event": [_hook("x", ["echo"])],
            HookEvent.ISSUE_CREATE.value: [],
            HookEvent.ISSUE_UPDATE.value: [
                _hook("dup", ["", "arg"]),
                _hook("dup", ["/does/not/exist"]),
                _hook("path-ok", ["./scripts/ok.sh"], cwd="missing-cwd"),
            ],
        },
        after={
            HookEvent.ISSUE_SHOW.value: [
                _hook("no-path", ["definitely-not-on-path"]),
            ]
        },
    )

    monkeypatch.setattr(
        hooks, "_resolve_hook_configuration", lambda _root: (cfg, tmp_path)
    )
    monkeypatch.setattr(
        hooks.shutil,
        "which",
        lambda executable: "/usr/bin/echo" if executable == "echo" else None,
    )

    issues = hooks.validate_hooks(tmp_path)

    assert any("unknown event" in issue for issue in issues)
    assert any("empty hook list" in issue for issue in issues)
    assert any("duplicate hook id 'dup'" in issue for issue in issues)
    assert any("command is empty" in issue for issue in issues)
    assert any("command not found" in issue for issue in issues)
    assert any("is not on PATH" in issue for issue in issues)
    assert any("cwd" in issue and "does not exist" in issue for issue in issues)


def test_run_lifecycle_hooks_short_circuits_and_blocking_failure(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    # Globally disabled
    monkeypatch.setattr(hooks, "hooks_globally_disabled", lambda _no_hooks: True)
    hooks.run_lifecycle_hooks(
        root=tmp_path,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_CREATE,
        actor="dev",
        beads_mode=False,
    )

    monkeypatch.setattr(hooks, "hooks_globally_disabled", lambda _no_hooks: False)

    # Config disabled
    monkeypatch.setattr(
        hooks,
        "_resolve_hook_configuration",
        lambda _root: (_hooks_config(enabled=False), tmp_path),
    )
    hooks.run_lifecycle_hooks(
        root=tmp_path,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_CREATE,
        actor="dev",
        beads_mode=False,
    )

    # Beads mode blocked
    monkeypatch.setattr(
        hooks,
        "_resolve_hook_configuration",
        lambda _root: (_hooks_config(run_in_beads_mode=False), tmp_path),
    )
    hooks.run_lifecycle_hooks(
        root=tmp_path,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_CREATE,
        actor="dev",
        beads_mode=True,
    )

    # Blocking failure on before/mutating event raises.
    cfg = _hooks_config(before={HookEvent.ISSUE_CREATE.value: [_hook("b1", ["echo"])]})
    monkeypatch.setattr(
        hooks, "_resolve_hook_configuration", lambda _root: (cfg, tmp_path)
    )

    class FailHandler:
        def run(self, **_kwargs):
            return HookResult(
                hook_id="b1",
                succeeded=False,
                timed_out=False,
                exit_code=2,
                duration_ms=1,
                message="failed",
            )

    monkeypatch.setattr(hooks, "ExternalCommandHookHandler", lambda: FailHandler())

    with pytest.raises(HookExecutionError, match="blocking hook 'b1' failed"):
        hooks.run_lifecycle_hooks(
            root=tmp_path,
            phase=HookPhase.BEFORE,
            event=HookEvent.ISSUE_CREATE,
            actor="dev",
            beads_mode=False,
        )


def test_run_lifecycle_hooks_nonblocking_warning_and_after_policy_provider(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    before_cfg = _hooks_config(
        before={HookEvent.ISSUE_SHOW.value: [_hook("s1", ["echo"], blocking=False)]}
    )
    monkeypatch.setattr(
        hooks, "_resolve_hook_configuration", lambda _root: (before_cfg, tmp_path)
    )

    class WarnHandler:
        def run(self, **_kwargs):
            return HookResult(
                hook_id="s1",
                succeeded=False,
                timed_out=False,
                exit_code=1,
                duration_ms=1,
                message="warn",
            )

    monkeypatch.setattr(hooks, "ExternalCommandHookHandler", lambda: WarnHandler())
    warned: list[str] = []
    monkeypatch.setattr(hooks.click, "echo", lambda msg, err=True: warned.append(msg))

    hooks.run_lifecycle_hooks(
        root=tmp_path,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_SHOW,
        actor="dev",
        beads_mode=False,
    )
    assert any("Hook warning (issue.show/before/s1): warn" in line for line in warned)

    # AFTER phase always invokes policy provider hook.
    after_cfg = _hooks_config(
        after={HookEvent.ISSUE_UPDATE.value: [_hook("a1", ["echo"])]}
    )
    monkeypatch.setattr(
        hooks, "_resolve_hook_configuration", lambda _root: (after_cfg, tmp_path)
    )

    class OkHandler:
        def run(self, **_kwargs):
            return HookResult(
                hook_id="a1",
                succeeded=True,
                timed_out=False,
                exit_code=0,
                duration_ms=1,
                message="ok",
            )

    monkeypatch.setattr(hooks, "ExternalCommandHookHandler", lambda: OkHandler())
    called: list[tuple[Path, HookEvent, bool, bool, int]] = []

    def fake_provider(
        *,
        project_root: Path,
        event: HookEvent,
        issues_for_policy,
        beads_mode: bool,
        no_guidance: bool,
    ) -> None:
        called.append(
            (project_root, event, beads_mode, no_guidance, len(issues_for_policy))
        )

    monkeypatch.setattr(hooks, "_run_policy_guidance_provider", fake_provider)

    hooks.run_lifecycle_hooks(
        root=tmp_path,
        phase=HookPhase.AFTER,
        event=HookEvent.ISSUE_UPDATE,
        actor="dev",
        beads_mode=False,
        issues_for_policy=[build_issue("kanbus-1")],
        no_guidance=True,
    )
    assert called == [(tmp_path, HookEvent.ISSUE_UPDATE, False, True, 1)]


def test_run_policy_guidance_provider_all_paths(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    issue = build_issue("kanbus-1")

    # No-op cases.
    hooks._run_policy_guidance_provider(
        project_root=Path("/repo"),
        event=HookEvent.ISSUE_UPDATE,
        issues_for_policy=[issue],
        beads_mode=True,
        no_guidance=False,
    )
    hooks._run_policy_guidance_provider(
        project_root=Path("/repo"),
        event=HookEvent.ISSUE_UPDATE,
        issues_for_policy=[],
        beads_mode=False,
        no_guidance=False,
    )
    hooks._run_policy_guidance_provider(
        project_root=Path("/repo"),
        event=HookEvent.ISSUE_COMMENT,
        issues_for_policy=[issue],
        beads_mode=False,
        no_guidance=False,
    )

    calls: list[tuple[str, int, bool]] = []

    def fake_emit(
        project_root: Path, issues_for_policy, operation, *, no_guidance: bool
    ) -> None:
        calls.append((str(project_root), len(issues_for_policy), no_guidance))

    monkeypatch.setattr("kanbus.policy_guidance.emit_guidance_for_issues", fake_emit)

    hooks._run_policy_guidance_provider(
        project_root=Path("/repo"),
        event=HookEvent.ISSUE_CREATE,
        issues_for_policy=[issue],
        beads_mode=False,
        no_guidance=True,
    )
    assert calls == [("/repo", 1, True)]

    monkeypatch.setattr(
        "kanbus.policy_guidance.emit_guidance_for_issues",
        lambda *_args, **_kwargs: (_ for _ in ()).throw(RuntimeError("guidance boom")),
    )
    warned: list[str] = []
    monkeypatch.setattr(hooks.click, "echo", lambda msg, err=True: warned.append(msg))

    hooks._run_policy_guidance_provider(
        project_root=Path("/repo"),
        event=HookEvent.ISSUE_UPDATE,
        issues_for_policy=[issue],
        beads_mode=False,
        no_guidance=False,
    )
    assert any(
        "Hook warning (issue.update/after/policy-guidance): guidance boom" in line
        for line in warned
    )


def test_parse_and_path_and_now_helpers() -> None:
    assert hooks._looks_like_path("./x") is True
    assert hooks._looks_like_path("dir/x") is True
    assert hooks._looks_like_path("dir\\x") is True
    assert hooks._looks_like_path("echo") is False

    assert hooks._parse_hook_event("issue.create") == HookEvent.ISSUE_CREATE
    assert hooks._parse_hook_event("invalid") is None

    now = hooks._now_utc_iso()
    assert now.endswith("Z")
    assert "T" in now


def test_validate_hooks_path_executable_branch(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    # Cover branch where executable is a path and exists.
    script = tmp_path / "bin" / "hook.sh"
    script.parent.mkdir(parents=True)
    script.write_text("echo ok", encoding="utf-8")

    cfg = _hooks_config(
        before={HookEvent.ISSUE_CREATE.value: [_hook("ok", [str(script)])]}
    )
    monkeypatch.setattr(
        hooks, "_resolve_hook_configuration", lambda _root: (cfg, tmp_path)
    )

    issues = hooks.validate_hooks(tmp_path)
    assert issues == []


def test_external_handler_absolute_cwd(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    handler = hooks.ExternalCommandHookHandler()
    invocation = hooks.HookInvocation(
        schema_version=hooks.HOOK_SCHEMA_VERSION,
        phase=HookPhase.BEFORE,
        event=HookEvent.ISSUE_UPDATE,
        timestamp="2026-03-09T00:00:00.000Z",
        actor="dev",
        mode={"runtime": "python"},
        operation={},
    )
    abs_cwd = tmp_path / "abs"
    abs_cwd.mkdir(parents=True)
    seen_cwd: list[Path] = []

    def ok_run(*_args, **kwargs):
        seen_cwd.append(Path(kwargs["cwd"]))
        return subprocess.CompletedProcess(
            args=["cmd"], returncode=0, stdout="", stderr=""
        )

    monkeypatch.setattr(hooks.subprocess, "run", ok_run)
    hook = _hook("x", ["cmd"], cwd=str(abs_cwd))
    result = handler.run(
        hook=hook, invocation=invocation, project_root=tmp_path, timeout_ms=1
    )
    assert result.succeeded is True
    assert seen_cwd == [abs_cwd]



"""Project discovery utilities."""

from __future__ import annotations

import os
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, List, Optional

from kanbus.config_loader import ConfigurationError, load_project_configuration
from kanbus.models import ProjectConfiguration


@dataclass
class ResolvedProject:
    """A resolved project directory with its label."""

    label: str
    project_dir: Path


class ProjectMarkerError(RuntimeError):
    """Raised when project discovery fails."""


WORKSPACE_IGNORE_DIRS = {
    ".git",
    ".mypy_cache",
    ".pytest_cache",
    ".venv",
    "__pycache__",
    "build",
    "dist",
    "node_modules",
    "project",
    "project-local",
    "target",
}

LEGACY_DISCOVERY_IGNORE_DIRS = {
    ".git",
    ".mypy_cache",
    ".pytest_cache",
    ".venv",
    "__pycache__",
    "build",
    "dist",
    "node_modules",
    "project-local",
    "target",
}


def discover_project_directories(root: Path) -> List[Path]:
    """Discover project directories beneath the current root.

    :param root: Root directory to search from.
    :type root: Path
    :return: List of discovered project directories.
    :rtype: List[Path]
    :raises ProjectMarkerError: If a configured project path is missing.
    """
    config_path = root / ".kanbus.yml"
    try:
        has_config = config_path.is_file()
    except OSError as error:
        raise ProjectMarkerError(str(error)) from error
    if has_config:
        project_dirs = _resolve_project_directories_from_config(config_path)
        for project_dir in project_dirs:
            if not project_dir.is_dir():
                raise ProjectMarkerError(f"kanbus path not found: {project_dir}")
        return _normalize_project_directories(project_dirs)

    project_dirs: List[Path] = []
    for workspace_config in _discover_workspace_config_paths(root):
        try:
            resolved = _resolve_project_directories_from_config(workspace_config)
        except ProjectMarkerError:
            continue
        for project_dir in resolved:
            if project_dir.is_dir():
                project_dirs.append(project_dir)
    if not project_dirs:
        project_dirs.extend(_discover_legacy_project_directories(root))
    return _normalize_project_directories(project_dirs)


def discover_kanbus_projects(root: Path) -> List[Path]:
    """Discover project directories from Kanbus configuration only.

    :param root: Root directory to search from.
    :type root: Path
    :return: List of configured project directories.
    :rtype: List[Path]
    :raises ProjectMarkerError: If a referenced path is missing.
    """
    config_path = root / ".kanbus.yml"
    if not config_path.is_file():
        return []
    return _normalize_project_directories(
        _resolve_project_directories_from_config(config_path)
    )


def resolve_labeled_projects(root: Path) -> List[ResolvedProject]:
    """Resolve all labeled project directories from configuration.

    :param root: Repository root.
    :type root: Path
    :return: List of resolved projects with labels.
    :rtype: List[ResolvedProject]
    :raises ProjectMarkerError: If configuration or paths are invalid.
    """
    config_path = get_configuration_path(root)
    try:
        configuration = load_project_configuration(config_path)
    except RuntimeError as error:
        raise ProjectMarkerError(str(error)) from error
    return _resolve_labeled_project_directories(config_path.parent, configuration)


def _resolve_labeled_project_directories(
    base: Path, configuration: ProjectConfiguration
) -> List[ResolvedProject]:
    projects: List[ResolvedProject] = []
    primary = base / configuration.project_directory
    projects.append(
        ResolvedProject(label=configuration.project_key, project_dir=primary)
    )
    for label, vp in configuration.virtual_projects.items():
        candidate = Path(vp.path)
        if not candidate.is_absolute():
            candidate = base / candidate
        candidate = resolve_project_path(candidate)
        if not candidate.is_dir():
            raise ProjectMarkerError(f"virtual project path not found: {candidate}")
        projects.append(ResolvedProject(label=label, project_dir=candidate))
    return projects


def _resolve_project_directories(
    base: Path, configuration: ProjectConfiguration
) -> List[Path]:
    paths: List[Path] = []
    primary = base / configuration.project_directory
    paths.append(primary)
    for vp in configuration.virtual_projects.values():
        candidate = Path(vp.path)
        if not candidate.is_absolute():
            candidate = base / candidate
        candidate = resolve_project_path(candidate)
        if not candidate.is_dir():
            raise ProjectMarkerError(f"virtual project path not found: {candidate}")
        paths.append(candidate)
    return paths


def _resolve_project_directories_from_config(config_path: Path) -> List[Path]:
    try:
        configuration = _load_configuration(config_path)
    except RuntimeError as error:
        raise ProjectMarkerError(str(error)) from error
    project_dirs = _resolve_project_directories(config_path.parent, configuration)
    return _apply_ignore_paths_for_config(
        config_path.parent, configuration, project_dirs
    )


def _apply_ignore_paths_for_config(
    base: Path, configuration: ProjectConfiguration, project_dirs: List[Path]
) -> List[Path]:
    if not configuration.ignore_paths:
        return project_dirs
    ignored = set()
    for pattern in configuration.ignore_paths:
        ignore_path = base / pattern
        try:
            ignored.add(ignore_path.resolve())
        except OSError:
            pass
    return [path for path in project_dirs if path.resolve() not in ignored]


def _discover_workspace_config_paths(root: Path) -> List[Path]:
    configs: List[Path] = []
    stack = [root]
    while stack:
        current = stack.pop()
        try:
            entries = list(current.iterdir())
        except OSError as error:
            if current == root:
                raise ProjectMarkerError(str(error)) from error
            continue
        config_path = current / ".kanbus.yml"
        if config_path.is_file():
            configs.append(config_path)
        for entry in entries:
            if not entry.is_dir():
                continue
            if entry.is_symlink():
                continue
            if entry.name in WORKSPACE_IGNORE_DIRS:
                continue
            stack.append(entry)
    return sorted(configs, key=str)


def _discover_legacy_project_directories(root: Path) -> List[Path]:
    projects: List[Path] = []
    if root.is_dir() and root.name == "project":
        projects.append(root)
    stack = [root]
    while stack:
        current = stack.pop()
        try:
            entries = list(current.iterdir())
        except OSError as error:
            if current == root:
                raise ProjectMarkerError(str(error)) from error
            continue
        for entry in entries:
            if not entry.is_dir():
                continue
            if entry.is_symlink():
                continue
            if entry.name == "project":
                projects.append(entry)
                continue
            if entry.name in LEGACY_DISCOVERY_IGNORE_DIRS:
                continue
            stack.append(entry)
    return projects


def _normalize_project_directories(paths: Iterable[Path]) -> List[Path]:
    normalized: List[Path] = []
    seen: set[Path] = set()
    for path in paths:
        candidate = resolve_project_path(path)
        if candidate in seen:
            continue
        seen.add(candidate)
        normalized.append(candidate)
    return sorted(normalized, key=str)


def resolve_project_path(path: Path) -> Path:
    """Resolve a project path while tolerating filesystem errors.

    :param path: Path to resolve.
    :type path: Path
    :return: Resolved path or original path on failure.
    :rtype: Path
    """
    try:
        return _resolve_path(path)
    except OSError:
        return path


def _resolve_path(path: Path) -> Path:
    if os.getenv("KANBUS_TEST_CANONICALIZE_FAILURE"):
        raise OSError("forced canonicalize failure")
    return path.resolve()


def _find_configuration_file(root: Path) -> Optional[Path]:
    current: Optional[Path] = root
    while current is not None:
        candidate = current / ".kanbus.yml"
        if candidate.is_file():
            return candidate
        parent = current.parent
        if parent == current:
            break
        current = parent
    return None


def _load_configuration(marker: Path) -> ProjectConfiguration:
    return load_project_configuration(marker)


def load_project_directory(root: Path) -> Path:
    """Load the primary project directory for write operations.

    When a configuration file is found, derives the project directory directly
    from ``project_directory`` in the config. Virtual project directories are
    for reading and lookup only and are not returned here.

    :param root: Repository root path.
    :type root: Path
    :return: Path to the primary project directory.
    :rtype: Path
    :raises ProjectMarkerError: If no project is found or the configuration
        is invalid.
    """
    marker = _find_configuration_file(root)
    if marker is not None:
        try:
            configuration = _load_configuration(marker)
        except RuntimeError as error:
            raise ProjectMarkerError(str(error)) from error
        primary = marker.parent / configuration.project_directory
        filtered = _apply_ignore_paths_for_config(
            marker.parent, configuration, [primary]
        )
        if not filtered:
            raise ProjectMarkerError("project not initialized")
        return _normalize_project_directories(filtered)[0]

    project_dirs = discover_project_directories(root)
    if not project_dirs:
        raise ProjectMarkerError("project not initialized")
    if len(project_dirs) > 1:
        discovered = ", ".join(str(path) for path in project_dirs)
        raise ProjectMarkerError(
            f"multiple projects found: {discovered}. "
            "Run this command from a directory with a single project/"
        )
    return project_dirs[0]


def get_configuration_path(root: Path) -> Path:
    """Return the configuration file path.

    :param root: Repository root path.
    :type root: Path
    :return: Path to .kanbus.yml.
    :rtype: Path
    :raises ProjectMarkerError: If the configuration file is missing.
    :raises ConfigurationError: If configuration path lookup fails.
    """
    if os.getenv("KANBUS_TEST_CONFIGURATION_PATH_FAILURE"):
        raise ConfigurationError("configuration path lookup failed")
    marker = _find_configuration_file(root)
    if marker is None:
        raise ProjectMarkerError("project not initialized")
    return marker


def find_project_local_directory(project_dir: Path) -> Optional[Path]:
    """Find a sibling project-local directory for a project.

    :param project_dir: Shared project directory.
    :type project_dir: Path
    :return: Project-local directory if present.
    :rtype: Optional[Path]
    """
    local_dir = project_dir.parent / "project-local"
    if local_dir.is_dir():
        return local_dir
    return None


def ensure_project_local_directory(project_dir: Path) -> Path:
    """Ensure the project-local directory exists and is gitignored.

    :param project_dir: Shared project directory.
    :type project_dir: Path
    :return: Path to the project-local directory.
    :rtype: Path
    """
    local_dir = project_dir.parent / "project-local"
    issues_dir = local_dir / "issues"
    events_dir = local_dir / "events"
    issues_dir.mkdir(parents=True, exist_ok=True)
    events_dir.mkdir(parents=True, exist_ok=True)
    _ensure_gitignore_entry(project_dir.parent, "project-local/")
    return local_dir


def _ensure_gitignore_entry(root: Path, entry: str) -> None:
    gitignore_path = root / ".gitignore"
    if gitignore_path.exists():
        contents = gitignore_path.read_text(encoding="utf-8")
        lines = [line.strip() for line in contents.splitlines()]
    else:
        contents = ""
        lines = []
    if entry in lines:
        return
    suffix = "" if contents.endswith("\n") or contents == "" else "\n"
    gitignore_path.write_text(f"{contents}{suffix}{entry}\n", encoding="utf-8")



from __future__ import annotations

import pytest

from kanbus import rich_text_signals


@pytest.mark.parametrize(
    "text",
    [
        "# Heading",
        "**bold**",
        "*italic*",
        "_em_",
        "```\ncode\n```",
        "`inline`",
        "> quote",
        "- list",
        "1. ordered",
        "---",
        "[link](https://example.com)",
    ],
)
def test_has_markdown_formatting_patterns(text: str) -> None:
    assert rich_text_signals.has_markdown_formatting(text) is True


def test_has_markdown_formatting_false_for_plain_text() -> None:
    assert rich_text_signals.has_markdown_formatting("plain text") is False


@pytest.mark.parametrize("language", ["mermaid", "plantuml", "d2", "Mermaid"])
def test_has_diagram_block_detects_supported_languages(language: str) -> None:
    text = f"```{language}\nA -> B\n```"
    assert rich_text_signals.has_diagram_block(text) is True


def test_has_diagram_block_false_without_supported_fence() -> None:
    assert rich_text_signals.has_diagram_block("```python\nprint(1)\n```") is False


def test_repair_escape_sequences_repairs_literal_newline_sequences() -> None:
    repaired, changed = rich_text_signals.repair_escape_sequences("one\\ntwo")
    assert changed is True
    assert repaired == "one\ntwo"


def test_repair_escape_sequences_returns_original_when_no_change() -> None:
    repaired, changed = rich_text_signals.repair_escape_sequences("one\ntwo")
    assert changed is False
    assert repaired == "one\ntwo"


def test_apply_text_quality_signals_repairs_and_emits_suggestions() -> None:
    result = rich_text_signals.apply_text_quality_signals("line1\\nline2")
    assert result.escape_sequences_repaired is True
    assert result.text == "line1\nline2"
    assert len(result.warnings) == 1
    assert "Literal \\n escape sequences" in result.warnings[0]
    assert len(result.suggestions) == 2


def test_apply_text_quality_signals_skips_suggestions_when_markdown_and_diagram_present() -> (
    None
):
    text = "# H\n```mermaid\nA-->B\n```"
    result = rich_text_signals.apply_text_quality_signals(text)
    assert result.escape_sequences_repaired is False
    assert result.warnings == []
    assert result.suggestions == []


def test_emit_signals_prints_warnings_suggestions_and_comment_hint(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    result = rich_text_signals.TextQualityResult(
        text="t",
        warnings=["w1"],
        suggestions=["s1", "s2"],
    )

    lines: list[str] = []
    monkeypatch.setattr(
        rich_text_signals.click, "echo", lambda msg, err=True: lines.append(msg)
    )

    rich_text_signals.emit_signals(
        result,
        context="description",
        issue_id="kanbus-1",
        comment_id="c1",
        is_update=True,
    )

    assert lines[0] == "w1"
    assert "s1" in lines[1]
    assert "s2" in lines[2]
    assert "kbs comment update kanbus-1 c1" in lines[3]


def test_emit_signals_without_suggestions_has_no_hint(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    result = rich_text_signals.TextQualityResult(
        text="t", warnings=["w"], suggestions=[]
    )
    lines: list[str] = []
    monkeypatch.setattr(
        rich_text_signals.click, "echo", lambda msg, err=True: lines.append(msg)
    )

    rich_text_signals.emit_signals(result, context="description", issue_id="kanbus-1")

    assert lines == ["w"]


def test_emit_follow_up_hint_uses_description_update_when_no_comment(
    monkeypatch: pytest.MonkeyPatch,
) -> None:
    lines: list[str] = []
    monkeypatch.setattr(
        rich_text_signals.click, "echo", lambda msg, err=True: lines.append(msg)
    )

    rich_text_signals._emit_follow_up_hint(
        context="description",
        issue_id="kanbus-2",
        comment_id=None,
        is_update=False,
    )

    assert lines == [
        '  -> To update the description: kbs update kanbus-2 --description "<your improved description here>"'
    ]



from __future__ import annotations

from datetime import datetime, timezone

from kanbus.models import IssueData, ProjectConfiguration


def build_issue(
    identifier: str,
    *,
    title: str = "Test issue",
    issue_type: str = "task",
    status: str = "open",
    priority: int = 2,
    parent: str | None = None,
    labels: list[str] | None = None,
    custom: dict[str, object] | None = None,
) -> IssueData:
    now = datetime(2026, 3, 6, tzinfo=timezone.utc)
    return IssueData.model_validate(
        {
            "id": identifier,
            "title": title,
            "description": "",
            "type": issue_type,
            "status": status,
            "priority": priority,
            "assignee": None,
            "creator": None,
            "parent": parent,
            "labels": labels or [],
            "dependencies": [],
            "comments": [],
            "created_at": now.isoformat(),
            "updated_at": now.isoformat(),
            "closed_at": None,
            "custom": custom or {},
        }
    )


def build_project_configuration(
    *,
    project_directory: str = "project",
    project_key: str = "kanbus",
    virtual_projects: dict[str, dict[str, str]] | None = None,
    beads_compatibility: bool = False,
) -> ProjectConfiguration:
    return ProjectConfiguration.model_validate(
        {
            "project_directory": project_directory,
            "virtual_projects": virtual_projects or {},
            "project_key": project_key,
            "hierarchy": ["initiative", "epic", "task", "sub-task"],
            "types": ["bug", "story", "chore"],
            "workflows": {
                "default": {
                    "open": ["in_progress", "closed"],
                    "in_progress": ["open", "closed"],
                    "closed": ["open"],
                }
            },
            "initial_status": "open",
            "priorities": {
                0: {"name": "critical", "color": "red"},
                1: {"name": "high", "color": "bright_red"},
                2: {"name": "medium", "color": "yellow"},
                3: {"name": "low", "color": "blue"},
                4: {"name": "trivial", "color": "white"},
            },
            "default_priority": 2,
            "beads_compatibility": beads_compatibility,
            "sort_order": {},
            "type_colors": {},
        }
    )



"""Issue deletion workflow."""

from __future__ import annotations

from pathlib import Path
from typing import List

from kanbus.issue_files import write_issue_to_file
from kanbus.issue_lookup import IssueLookupError, load_issue_from_project
from kanbus.event_history import (
    delete_events_for_issues,
    events_dir_for_issue_path,
)
from kanbus.issue_listing import load_issues_from_directory
from kanbus.project import find_project_local_directory
from kanbus.gossip import publish_issue_deleted


class IssueDeleteError(RuntimeError):
    """Raised when issue deletion fails."""


def get_descendant_identifiers(project_dir: Path, identifier: str) -> List[str]:
    """Return descendant issue identifiers in leaf-first order (children before parents).

    :param project_dir: Shared project directory.
    :type project_dir: Path
    :param identifier: Root issue identifier.
    :type identifier: str
    :return: List of descendant IDs, deepest first.
    :rtype: List[str]
    """
    parent_to_children: dict[str, List[str]] = {}
    issues_dir = project_dir / "issues"
    if issues_dir.is_dir():
        for issue in load_issues_from_directory(issues_dir):
            if issue.parent is not None:
                parent_to_children.setdefault(issue.parent, []).append(issue.identifier)
    local_dir = find_project_local_directory(project_dir)
    if local_dir is not None:
        local_issues_dir = local_dir / "issues"
        if local_issues_dir.is_dir():
            for issue in load_issues_from_directory(local_issues_dir):
                if issue.parent is not None:
                    parent_to_children.setdefault(issue.parent, []).append(
                        issue.identifier
                    )
    depth: dict[str, int] = {identifier: 0}
    queue: List[str] = [identifier]
    while queue:
        parent_id = queue.pop(0)
        for child_id in parent_to_children.get(parent_id, []):
            if child_id not in depth:
                depth[child_id] = depth[parent_id] + 1
                queue.append(child_id)
    descendants = [k for k in depth if k != identifier]
    return sorted(descendants, key=lambda x: -depth[x])


def delete_issue(root: Path, identifier: str) -> None:
    """Delete an issue file and all its event history from disk.

    :param root: Repository root path.
    :type root: Path
    :param identifier: Issue identifier.
    :type identifier: str
    :raises IssueDeleteError: If deletion fails.
    """
    try:
        lookup = load_issue_from_project(root, identifier)
    except IssueLookupError as error:
        raise IssueDeleteError(str(error)) from error

    issue_id = lookup.issue.identifier
    events_dir = events_dir_for_issue_path(lookup.project_dir, lookup.issue_path)
    lookup.issue_path.unlink()
    try:
        delete_events_for_issues(events_dir, {issue_id})
    except Exception as error:  # noqa: BLE001
        write_issue_to_file(lookup.issue, lookup.issue_path)
        raise IssueDeleteError(str(error)) from error
    if lookup.issue_path.parent == lookup.project_dir / "issues":
        publish_issue_deleted(root, lookup.project_dir, issue_id, None)



"""Index cache utilities for Kanbus."""

from __future__ import annotations

import json
import os
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import Dict, List, Optional

from kanbus.index import IssueIndex
from kanbus.models import IssueData


@dataclass
class IndexCache:
    """Serialized index cache representation."""

    version: int
    built_at: datetime
    file_mtimes: Dict[str, float]
    issues: List[IssueData]
    reverse_deps: Dict[str, List[str]]


def _normalize_mtime(value: float) -> float:
    """Normalize file modification times for cache consistency.

    :param value: Raw modification time value.
    :type value: float
    :return: Rounded modification time to microsecond precision.
    :rtype: float
    """
    return round(value, 6)


def collect_issue_file_mtimes(issues_directory: Path) -> Dict[str, float]:
    """Collect file modification times for issues.

    :param issues_directory: Directory containing issue files.
    :type issues_directory: Path
    :return: Mapping of filename to mtime.
    :rtype: Dict[str, float]
    """
    mtimes: Dict[str, float] = {}
    for entry in os.scandir(issues_directory):
        if not entry.is_file() or not entry.name.endswith(".json"):
            continue
        mtimes[entry.name] = _normalize_mtime(entry.stat().st_mtime)
    return mtimes


def load_cache_if_valid(
    cache_path: Path, issues_directory: Path
) -> Optional[IssueIndex]:
    """Load cached index if the cache is valid.

    :param cache_path: Path to cache file.
    :type cache_path: Path
    :param issues_directory: Directory containing issue files.
    :type issues_directory: Path
    :return: IssueIndex if cache is valid, otherwise None.
    :rtype: Optional[IssueIndex]
    """
    if not cache_path.exists():
        return None

    payload = json.loads(cache_path.read_text(encoding="utf-8"))
    file_mtimes = payload.get("file_mtimes", {})
    current_mtimes = collect_issue_file_mtimes(issues_directory)
    if file_mtimes != current_mtimes:
        return None

    issues = [IssueData.model_validate(item) for item in payload.get("issues", [])]
    reverse_deps = payload.get("reverse_deps", {})
    return build_index_from_cache(issues, reverse_deps)


def write_cache(
    index: IssueIndex, cache_path: Path, file_mtimes: Dict[str, float]
) -> None:
    """Write an index cache file to disk.

    :param index: Issue index to serialize.
    :type index: IssueIndex
    :param cache_path: Path to cache file.
    :type cache_path: Path
    :param file_mtimes: File modification time mapping.
    :type file_mtimes: Dict[str, float]
    """
    cache = IndexCache(
        version=1,
        built_at=datetime.now(timezone.utc),
        file_mtimes=file_mtimes,
        issues=list(index.by_id.values()),
        reverse_deps={
            target: [issue.identifier for issue in issues]
            for target, issues in index.reverse_dependencies.items()
        },
    )
    payload = {
        "version": cache.version,
        "built_at": cache.built_at.isoformat().replace("+00:00", "Z"),
        "file_mtimes": cache.file_mtimes,
        "issues": [
            issue.model_dump(by_alias=True, mode="json") for issue in cache.issues
        ],
        "reverse_deps": cache.reverse_deps,
    }
    cache_path.parent.mkdir(parents=True, exist_ok=True)
    cache_path.write_text(
        json.dumps(payload, indent=2, sort_keys=False),
        encoding="utf-8",
    )


def build_index_from_cache(
    issues: List[IssueData], reverse_deps: Dict[str, List[str]]
) -> IssueIndex:
    """Rebuild an IssueIndex from cached data.

    :param issues: Issues from cache.
    :type issues: List[IssueData]
    :param reverse_deps: Reverse dependency mapping.
    :type reverse_deps: Dict[str, List[str]]
    :return: IssueIndex populated from cache.
    :rtype: IssueIndex
    """
    index = IssueIndex()
    for issue in issues:
        index.by_id[issue.identifier] = issue
        index.by_status.setdefault(issue.status, []).append(issue)
        index.by_type.setdefault(issue.issue_type, []).append(issue)
        if issue.parent is not None:
            index.by_parent.setdefault(issue.parent, []).append(issue)
        for label in issue.labels:
            index.by_label.setdefault(label, []).append(issue)

    for target, issue_ids in reverse_deps.items():
        index.reverse_dependencies[target] = [
            index.by_id[identifier]
            for identifier in issue_ids
            if identifier in index.by_id
        ]

    return index



from __future__ import annotations

from pathlib import Path
from types import SimpleNamespace

from click.testing import CliRunner
import pytest

from kanbus import cli
from kanbus.config_loader import ConfigurationError
from kanbus.issue_delete import IssueDeleteError
from kanbus.issue_listing import IssueListingError
from kanbus.issue_lookup import IssueLookupError
from kanbus.issue_update import IssueUpdateError
from kanbus.queries import QueryError

from test_helpers import build_issue, build_project_configuration


def _run(args: list[str]) -> object:
    return CliRunner().invoke(cli.cli, args)


def test_bulk_update_paths(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)

    result_beads = _run(
        ["--beads", "bulk", "update", "--id", "kanbus-1", "--set-status", "done"]
    )
    assert result_beads.exit_code != 0
    assert "not supported in beads mode" in result_beads.output

    result_missing_selector = _run(["bulk", "update", "--set-status", "done"])
    assert result_missing_selector.exit_code != 0
    assert "requires at least one selector" in result_missing_selector.output

    result_missing_setter = _run(["bulk", "update", "--id", "kanbus-1"])
    assert result_missing_setter.exit_code != 0
    assert "requires at least one setter" in result_missing_setter.output

    update_calls: list[tuple[str, bool]] = []

    def _update_issue(**kwargs):
        identifier = kwargs["identifier"]
        update_calls.append((identifier, kwargs["validate"]))
        return build_issue(identifier)

    monkeypatch.setattr(cli, "update_issue", _update_issue)
    monkeypatch.setattr(
        cli,
        "list_issues",
        lambda *_a, **_k: [build_issue("kanbus-1"), build_issue("kanbus-2")],
    )

    result_success = _run(
        [
            "bulk",
            "update",
            "--id",
            "kanbus-1",
            "--where-status",
            "open",
            "--set-status",
            "done",
            "--set-assignee",
            "ryan",
            "--no-validate",
        ]
    )
    assert result_success.exit_code == 0
    assert "Updated 2 issue(s)" in result_success.output
    assert update_calls == [("kanbus-1", False), ("kanbus-2", False)]

    monkeypatch.setattr(
        cli,
        "update_issue",
        lambda **_k: (_ for _ in ()).throw(IssueUpdateError("update fail")),
    )
    result_update_fail = _run(
        ["bulk", "update", "--id", "kanbus-1", "--set-status", "done"]
    )
    assert result_update_fail.exit_code != 0
    assert "update fail" in result_update_fail.output

    monkeypatch.setattr(cli, "update_issue", _update_issue)
    monkeypatch.setattr(
        cli,
        "list_issues",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueListingError("list fail")),
    )
    result_list_fail = _run(
        ["bulk", "update", "--where-status", "open", "--set-status", "done"]
    )
    assert result_list_fail.exit_code != 0
    assert "list fail" in result_list_fail.output

    call_order = {"count": 0}

    def _update_issue_second_fails(**kwargs):
        call_order["count"] += 1
        if call_order["count"] == 2:
            raise IssueUpdateError("second update fail")
        return build_issue(kwargs["identifier"])

    monkeypatch.setattr(cli, "update_issue", _update_issue_second_fails)
    monkeypatch.setattr(cli, "list_issues", lambda *_a, **_k: [build_issue("kanbus-2")])
    result_second_update_fail = _run(
        [
            "bulk",
            "update",
            "--id",
            "kanbus-1",
            "--where-status",
            "open",
            "--set-status",
            "done",
        ]
    )
    assert result_second_update_fail.exit_code != 0
    assert "second update fail" in result_second_update_fail.output


def test_delete_paths_regular_mode(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(cli.ProjectMarkerError("pm")),
    )
    monkeypatch.setattr(
        cli,
        "format_issue_key",
        lambda identifier, project_context=False: identifier,
    )

    monkeypatch.setattr(cli, "_delete_terminal_is_interactive", lambda: False)
    result_non_interactive = _run(["delete", "kanbus-1"])
    assert result_non_interactive.exit_code != 0
    assert "requires confirmation" in result_non_interactive.output

    issue = build_issue("kanbus-1")
    monkeypatch.setattr(cli, "_delete_terminal_is_interactive", lambda: True)
    monkeypatch.setattr(cli.click, "confirm", lambda *_a, **_k: False)
    monkeypatch.setattr(
        cli,
        "load_issue_from_project",
        lambda *_a, **_k: SimpleNamespace(issue=issue, project_dir=tmp_path / "proj"),
    )
    result_cancel = _run(["delete", "kanbus-1"])
    assert result_cancel.exit_code == 0
    assert "Delete cancelled." in result_cancel.output

    confirmations = iter([True, False])
    monkeypatch.setattr(cli.click, "confirm", lambda *_a, **_k: next(confirmations))
    monkeypatch.setattr(
        cli,
        "get_descendant_identifiers",
        lambda *_a, **_k: ["d1", "d2", "d3", "d4", "d5", "d6"],
    )
    deleted: list[str] = []
    monkeypatch.setattr(
        cli, "delete_issue", lambda _root, issue_id: deleted.append(issue_id)
    )

    result_recursive = _run(["delete", "kanbus-1", "--recursive"])
    assert result_recursive.exit_code == 0
    assert deleted == ["kanbus-1"]

    monkeypatch.setattr(
        cli,
        "load_issue_from_project",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueLookupError("lookup fail")),
    )
    result_lookup_fail = _run(["delete", "kanbus-1", "--yes"])
    assert result_lookup_fail.exit_code != 0
    assert "lookup fail" in result_lookup_fail.output

    monkeypatch.setattr(
        cli,
        "load_issue_from_project",
        lambda *_a, **_k: SimpleNamespace(issue=issue, project_dir=tmp_path / "proj"),
    )
    monkeypatch.setattr(
        cli,
        "delete_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueDeleteError("delete fail")),
    )
    result_delete_fail = _run(["delete", "kanbus-1", "--yes"])
    assert result_delete_fail.exit_code != 0
    assert "delete fail" in result_delete_fail.output


def test_delete_paths_beads_mode(
    monkeypatch: pytest.MonkeyPatch, tmp_path: Path
) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)
    monkeypatch.setattr(cli, "_resolve_beads_root", lambda _cwd: tmp_path)
    monkeypatch.setattr(
        cli, "format_issue_key", lambda identifier, project_context=False: identifier
    )
    monkeypatch.setattr(
        cli, "load_beads_issue", lambda *_a, **_k: build_issue("kanbus-1")
    )

    monkeypatch.setattr(cli, "_delete_terminal_is_interactive", lambda: False)
    result_non_interactive = _run(["--beads", "delete", "kanbus-1"])
    assert result_non_interactive.exit_code != 0
    assert "requires confirmation" in result_non_interactive.output

    monkeypatch.setattr(cli, "_delete_terminal_is_interactive", lambda: True)
    monkeypatch.setattr(cli.click, "confirm", lambda *_a, **_k: False)
    result_cancel = _run(["--beads", "delete", "kanbus-1"])
    assert result_cancel.exit_code == 0
    assert "Delete cancelled." in result_cancel.output

    confirmations = iter([True, False])
    monkeypatch.setattr(cli.click, "confirm", lambda *_a, **_k: next(confirmations))
    monkeypatch.setattr(
        cli,
        "get_beads_descendant_identifiers",
        lambda *_a, **_k: ["a", "b", "c", "d", "e", "f"],
    )
    recursive_flags: list[bool] = []
    monkeypatch.setattr(
        cli,
        "delete_beads_issue",
        lambda _root, _identifier, recursive: recursive_flags.append(recursive),
    )
    result_recursive = _run(["--beads", "delete", "kanbus-1", "--recursive"])
    assert result_recursive.exit_code == 0
    assert recursive_flags == [False]

    monkeypatch.setattr(
        cli,
        "delete_beads_issue",
        lambda *_a, **_k: (_ for _ in ()).throw(cli.BeadsDeleteError("beads fail")),
    )
    result_fail = _run(["--beads", "delete", "kanbus-1", "--yes"])
    assert result_fail.exit_code != 0
    assert "beads fail" in result_fail.output


def test_list_command_paths(monkeypatch: pytest.MonkeyPatch, tmp_path: Path) -> None:
    monkeypatch.setattr(cli.Path, "cwd", lambda: tmp_path)
    monkeypatch.setattr(cli, "_run_lifecycle_hooks_for_context", lambda *_a, **_k: None)

    regular_issue_a = build_issue("kanbus-1")
    regular_issue_b = build_issue("kanbus-2", custom={"project_path": "proj-b"})

    widths_calls: list[bool] = []
    monkeypatch.setattr(
        cli,
        "compute_widths",
        lambda issues, project_context: widths_calls.append(project_context)
        or {"id": 8},
    )
    monkeypatch.setattr(
        cli,
        "format_issue_line",
        lambda issue, porcelain, widths, project_context, configuration: (
            f"{issue.identifier}:{porcelain}:{project_context}:{configuration is not None}"
        ),
    )
    monkeypatch.setattr(
        cli,
        "list_issues",
        lambda *_a, **_k: [regular_issue_a, regular_issue_b],
    )
    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(),
    )
    monkeypatch.setattr(
        cli, "get_configuration_path", lambda _p: tmp_path / ".kanbus.yml"
    )

    result_regular = _run(["list", "--limit", "1", "--full-ids"])
    assert result_regular.exit_code == 0
    assert "kanbus-1:False:False:True" in result_regular.output
    assert widths_calls == [False]

    widths_calls.clear()
    monkeypatch.setattr(cli, "list_issues", lambda *_a, **_k: [regular_issue_a])
    result_project_context = _run(["list", "--limit", "1"])
    assert result_project_context.exit_code == 0
    assert "kanbus-1:False:True:True" in result_project_context.output
    assert widths_calls == [True]

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: (_ for _ in ()).throw(ConfigurationError("config fail")),
    )
    result_config_fail = _run(["list"])
    assert result_config_fail.exit_code != 0
    assert "config fail" in result_config_fail.output

    monkeypatch.setattr(
        cli,
        "load_project_configuration",
        lambda _p: build_project_configuration(),
    )
    monkeypatch.setattr(
        cli,
        "list_issues",
        lambda *_a, **_k: (_ for _ in ()).throw(IssueListingError("list fail")),
    )
    result_list_fail = _run(["list"])
    assert result_list_fail.exit_code != 0
    assert "list fail" in result_list_fail.output

    monkeypatch.setattr(
        cli,
        "list_issues",
        lambda *_a, **_k: (_ for _ in ()).throw(QueryError("query fail")),
    )
    result_query_fail = _run(["list"])
    assert result_query_fail.exit_code != 0
    assert "query fail" in result_query_fail.output

    beads_issue_a = build_issue("kanbus-3", priority=3)
    beads_issue_b = build_issue("kanbus-2", priority=1)
    monkeypatch.setattr(cli, "_resolve_beads_root", lambda _cwd: tmp_path)
    monkeypatch.setattr(
        cli, "list_issues", lambda *_a, **_k: [beads_issue_a, beads_issue_b]
    )
    monkeypatch.setattr(
        cli,
        "compute_widths",
        lambda *_a, **_k: (_ for _ in ()).throw(AssertionError("unexpected widths")),
    )
    monkeypatch.setattr(
        cli,
        "format_issue_line",
        lambda issue, porcelain, widths, project_context, configuration: f"{issue.identifier}:{porcelain}:{project_context}:{configuration}",
    )

    result_beads = _run(["--beads", "list", "--porcelain", "--limit", "0"])
    assert result_beads.exit_code == 0
    lines = [line for line in result_beads.output.splitlines() if line.strip()]
    assert lines == [
        "kanbus-2:True:False:None",
        "kanbus-3:True:False:None",
    ]



from __future__ import annotations

from pathlib import Path

import pytest

from kanbus import project
from kanbus.config_loader import ConfigurationError

from test_helpers import build_project_configuration


def test_discover_project_directories_from_root_config(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    (tmp_path / ".kanbus.yml").write_text("x", encoding="utf-8")
    p1 = tmp_path / "project"
    p1.mkdir()
    monkeypatch.setattr(
        project, "_resolve_project_directories_from_config", lambda _cfg: [p1]
    )

    discovered = project.discover_project_directories(tmp_path)
    assert discovered == [p1.resolve()]


def test_discover_project_directories_wraps_is_file_oserror(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    original_is_file = Path.is_file
    config_path = tmp_path / ".kanbus.yml"

    def fake_is_file(path_self: Path) -> bool:
        if path_self == config_path:
            raise OSError("cannot stat config")
        return original_is_file(path_self)

    monkeypatch.setattr(Path, "is_file", fake_is_file)
    with pytest.raises(project.ProjectMarkerError, match="cannot stat config"):
        project.discover_project_directories(tmp_path)


def test_discover_project_directories_raises_for_missing_configured_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    (tmp_path / ".kanbus.yml").write_text("x", encoding="utf-8")
    missing = tmp_path / "missing"
    monkeypatch.setattr(
        project, "_resolve_project_directories_from_config", lambda _cfg: [missing]
    )

    with pytest.raises(project.ProjectMarkerError, match="kanbus path not found"):
        project.discover_project_directories(tmp_path)


def test_discover_project_directories_workspace_config_fallback(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    workspace = tmp_path / "workspace"
    workspace.mkdir()
    cfg = workspace / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    proj = workspace / "project"
    proj.mkdir()

    monkeypatch.setattr(
        project, "_discover_workspace_config_paths", lambda _root: [cfg]
    )
    monkeypatch.setattr(
        project, "_resolve_project_directories_from_config", lambda _cfg: [proj]
    )

    discovered = project.discover_project_directories(tmp_path)
    assert discovered == [proj.resolve()]


def test_discover_project_directories_handles_workspace_config_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / "workspace" / ".kanbus.yml"
    monkeypatch.setattr(
        project, "_discover_workspace_config_paths", lambda _root: [cfg]
    )
    monkeypatch.setattr(
        project,
        "_resolve_project_directories_from_config",
        lambda _cfg: (_ for _ in ()).throw(project.ProjectMarkerError("bad config")),
    )
    monkeypatch.setattr(
        project, "_discover_legacy_project_directories", lambda _root: []
    )

    assert project.discover_project_directories(tmp_path) == []


def test_discover_project_directories_uses_legacy_when_no_workspace_results(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    legacy = tmp_path / "nested" / "project"
    legacy.mkdir(parents=True)
    monkeypatch.setattr(project, "_discover_workspace_config_paths", lambda _root: [])
    monkeypatch.setattr(
        project, "_discover_legacy_project_directories", lambda _root: [legacy]
    )

    discovered = project.discover_project_directories(tmp_path)
    assert discovered == [legacy.resolve()]


def test_discover_kanbus_projects_empty_without_config(tmp_path: Path) -> None:
    assert project.discover_kanbus_projects(tmp_path) == []


def test_discover_kanbus_projects_from_root_config(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    p1 = tmp_path / "project"
    p1.mkdir()
    monkeypatch.setattr(
        project, "_resolve_project_directories_from_config", lambda _cfg: [p1]
    )

    discovered = project.discover_kanbus_projects(tmp_path)
    assert discovered == [p1.resolve()]


def test_resolve_labeled_projects_success(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    primary = tmp_path / "project"
    primary.mkdir()
    virtual = tmp_path / "other"
    virtual.mkdir()
    configuration = build_project_configuration(
        project_directory="project", virtual_projects={"ops": {"path": "other"}}
    )

    monkeypatch.setattr(project, "get_configuration_path", lambda _root: cfg)
    monkeypatch.setattr(
        project, "load_project_configuration", lambda _cfg: configuration
    )

    resolved = project.resolve_labeled_projects(tmp_path)
    labels = [p.label for p in resolved]
    assert labels == [configuration.project_key, "ops"]


def test_resolve_labeled_projects_wraps_config_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    monkeypatch.setattr(project, "get_configuration_path", lambda _root: cfg)
    monkeypatch.setattr(
        project,
        "load_project_configuration",
        lambda _cfg: (_ for _ in ()).throw(RuntimeError("bad")),
    )

    with pytest.raises(project.ProjectMarkerError, match="bad"):
        project.resolve_labeled_projects(tmp_path)


def test_resolve_labeled_project_directories_raises_for_missing_virtual(
    tmp_path: Path,
) -> None:
    configuration = build_project_configuration(
        project_directory="project", virtual_projects={"ops": {"path": "missing"}}
    )

    with pytest.raises(
        project.ProjectMarkerError, match="virtual project path not found"
    ):
        project._resolve_labeled_project_directories(tmp_path, configuration)


def test_resolve_project_directories_raises_for_missing_virtual(tmp_path: Path) -> None:
    configuration = build_project_configuration(
        project_directory="project", virtual_projects={"ops": {"path": "missing"}}
    )

    with pytest.raises(
        project.ProjectMarkerError, match="virtual project path not found"
    ):
        project._resolve_project_directories(tmp_path, configuration)


def test_resolve_project_directories_includes_existing_virtual(tmp_path: Path) -> None:
    (tmp_path / "project").mkdir()
    (tmp_path / "vp").mkdir()
    configuration = build_project_configuration(
        project_directory="project", virtual_projects={"ops": {"path": "vp"}}
    )

    resolved = project._resolve_project_directories(tmp_path, configuration)
    assert resolved == [tmp_path / "project", (tmp_path / "vp").resolve()]


def test_resolve_project_directories_from_config_wraps_runtime(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    monkeypatch.setattr(
        project,
        "_load_configuration",
        lambda _cfg: (_ for _ in ()).throw(RuntimeError("oops")),
    )

    with pytest.raises(project.ProjectMarkerError, match="oops"):
        project._resolve_project_directories_from_config(cfg)


def test_resolve_project_directories_from_config_success(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    configuration = build_project_configuration()
    expected = [tmp_path / "project"]
    monkeypatch.setattr(project, "_load_configuration", lambda _cfg: configuration)
    monkeypatch.setattr(
        project, "_resolve_project_directories", lambda _base, _c: expected
    )

    resolved = project._resolve_project_directories_from_config(cfg)
    assert resolved == expected


def test_apply_ignore_paths_for_config(tmp_path: Path) -> None:
    keep = tmp_path / "project"
    drop = tmp_path / "drop"
    keep.mkdir()
    drop.mkdir()
    configuration = build_project_configuration(project_directory="project")
    configuration.ignore_paths = ["drop"]

    filtered = project._apply_ignore_paths_for_config(
        tmp_path, configuration, [keep, drop]
    )
    assert filtered == [keep]


def test_apply_ignore_paths_tolerates_unresolvable_ignore_path(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    keep = tmp_path / "project"
    keep.mkdir()
    configuration = build_project_configuration(project_directory="project")
    configuration.ignore_paths = ["broken"]
    original_resolve = Path.resolve
    broken = tmp_path / "broken"

    def fake_resolve(path_self: Path, *args, **kwargs):  # type: ignore[no-untyped-def]
        if path_self == broken:
            raise OSError("bad path")
        return original_resolve(path_self, *args, **kwargs)

    monkeypatch.setattr(Path, "resolve", fake_resolve)
    filtered = project._apply_ignore_paths_for_config(tmp_path, configuration, [keep])
    assert filtered == [keep]


def test_discover_workspace_config_paths_and_legacy_discovery(tmp_path: Path) -> None:
    (tmp_path / ".git").mkdir()
    work = tmp_path / "work"
    work.mkdir()
    (work / ".kanbus.yml").write_text("x", encoding="utf-8")
    nested = work / "nested"
    nested.mkdir()
    (nested / "project").mkdir()

    configs = project._discover_workspace_config_paths(tmp_path)
    legacy = project._discover_legacy_project_directories(tmp_path)

    assert configs == [work / ".kanbus.yml"]
    assert nested / "project" in legacy


def test_discover_workspace_config_paths_raises_on_root_oserror(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path / "root"
    root.mkdir()
    original_iterdir = Path.iterdir

    def fake_iterdir(path_self: Path):
        if path_self == root:
            raise OSError("cannot read root")
        return original_iterdir(path_self)

    monkeypatch.setattr(Path, "iterdir", fake_iterdir)

    with pytest.raises(project.ProjectMarkerError, match="cannot read root"):
        project._discover_workspace_config_paths(root)


def test_discover_workspace_config_paths_skips_non_root_oserror_and_symlink(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path / "root"
    root.mkdir()
    bad = root / "bad"
    bad.mkdir()
    (bad / ".kanbus.yml").write_text("x", encoding="utf-8")
    (root / "ok").mkdir()
    (root / "ok" / ".kanbus.yml").write_text("x", encoding="utf-8")
    (root / "sym").symlink_to(root / "ok", target_is_directory=True)
    original_iterdir = Path.iterdir

    def fake_iterdir(path_self: Path):
        if path_self == bad:
            raise OSError("cannot read nested")
        return original_iterdir(path_self)

    monkeypatch.setattr(Path, "iterdir", fake_iterdir)
    configs = project._discover_workspace_config_paths(root)
    assert configs == [root / "ok" / ".kanbus.yml"]


def test_normalize_project_directories_deduplicates_and_sorts(tmp_path: Path) -> None:
    a = tmp_path / "a"
    b = tmp_path / "b"
    a.mkdir()
    b.mkdir()

    normalized = project._normalize_project_directories([b, a, b])
    assert normalized == [a.resolve(), b.resolve()]


def test_resolve_project_path_tolerates_forced_failure(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setenv("KANBUS_TEST_CANONICALIZE_FAILURE", "1")
    path = tmp_path / "x"
    assert project.resolve_project_path(path) == path


def test_find_configuration_file_walks_up(tmp_path: Path) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    nested = tmp_path / "a" / "b"
    nested.mkdir(parents=True)

    assert project._find_configuration_file(nested) == cfg


def test_find_configuration_file_returns_none_at_filesystem_root() -> None:
    assert project._find_configuration_file(Path("/")) is None


def test_load_project_directory_from_config_and_error_paths(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    primary = tmp_path / "project"
    primary.mkdir()
    configuration = build_project_configuration(project_directory="project")

    monkeypatch.setattr(project, "_find_configuration_file", lambda _root: cfg)
    monkeypatch.setattr(project, "_load_configuration", lambda _marker: configuration)

    assert project.load_project_directory(tmp_path) == primary.resolve()

    configuration.ignore_paths = ["project"]
    with pytest.raises(project.ProjectMarkerError, match="project not initialized"):
        project.load_project_directory(tmp_path)


def test_load_project_directory_wraps_configuration_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    monkeypatch.setattr(project, "_find_configuration_file", lambda _root: cfg)
    monkeypatch.setattr(
        project,
        "_load_configuration",
        lambda _marker: (_ for _ in ()).throw(RuntimeError("bad config")),
    )

    with pytest.raises(project.ProjectMarkerError, match="bad config"):
        project.load_project_directory(tmp_path)


def test_load_project_directory_discovery_errors(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    monkeypatch.setattr(project, "_find_configuration_file", lambda _root: None)
    monkeypatch.setattr(project, "discover_project_directories", lambda _root: [])
    with pytest.raises(project.ProjectMarkerError, match="project not initialized"):
        project.load_project_directory(tmp_path)

    p1 = tmp_path / "a"
    p2 = tmp_path / "b"
    p1.mkdir()
    p2.mkdir()
    monkeypatch.setattr(project, "discover_project_directories", lambda _root: [p1, p2])
    with pytest.raises(project.ProjectMarkerError, match="multiple projects found"):
        project.load_project_directory(tmp_path)


def test_load_project_directory_discovery_single_result(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    p1 = tmp_path / "project"
    p1.mkdir()
    monkeypatch.setattr(project, "_find_configuration_file", lambda _root: None)
    monkeypatch.setattr(project, "discover_project_directories", lambda _root: [p1])
    assert project.load_project_directory(tmp_path) == p1


def test_get_configuration_path_variants(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")

    monkeypatch.setenv("KANBUS_TEST_CONFIGURATION_PATH_FAILURE", "1")
    with pytest.raises(ConfigurationError, match="configuration path lookup failed"):
        project.get_configuration_path(tmp_path)

    monkeypatch.delenv("KANBUS_TEST_CONFIGURATION_PATH_FAILURE", raising=False)
    monkeypatch.setattr(project, "_find_configuration_file", lambda _root: None)
    with pytest.raises(project.ProjectMarkerError, match="project not initialized"):
        project.get_configuration_path(tmp_path)

    monkeypatch.setattr(project, "_find_configuration_file", lambda _root: cfg)
    assert project.get_configuration_path(tmp_path) == cfg


def test_project_local_helpers(tmp_path: Path) -> None:
    project_dir = tmp_path / "project"
    project_dir.mkdir()

    assert project.find_project_local_directory(project_dir) is None

    local_dir = project.ensure_project_local_directory(project_dir)
    assert local_dir == tmp_path / "project-local"
    assert (local_dir / "issues").is_dir()
    assert (local_dir / "events").is_dir()
    assert project.find_project_local_directory(project_dir) == local_dir

    gitignore = tmp_path / ".gitignore"
    contents = gitignore.read_text(encoding="utf-8")
    assert "project-local/" in contents


def test_ensure_gitignore_entry_is_idempotent(tmp_path: Path) -> None:
    gitignore = tmp_path / ".gitignore"
    gitignore.write_text("foo\nproject-local/\n", encoding="utf-8")

    project._ensure_gitignore_entry(tmp_path, "project-local/")
    assert gitignore.read_text(encoding="utf-8").count("project-local/") == 1


def test_legacy_discovery_handles_root_named_project_and_root_oserror(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root_project = tmp_path / "project"
    root_project.mkdir()
    assert root_project in project._discover_legacy_project_directories(root_project)

    root = tmp_path / "broken"
    root.mkdir()
    original_iterdir = Path.iterdir

    def fake_iterdir(path_self: Path):
        if path_self == root:
            raise OSError("cannot read root")
        return original_iterdir(path_self)

    monkeypatch.setattr(Path, "iterdir", fake_iterdir)
    with pytest.raises(project.ProjectMarkerError, match="cannot read root"):
        project._discover_legacy_project_directories(root)


def test_legacy_discovery_skips_symlink_and_non_root_oserror(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    root = tmp_path / "root"
    root.mkdir()
    child = root / "child"
    child.mkdir()
    (child / "project").mkdir()
    bad = root / "bad"
    bad.mkdir()
    (root / "sym").symlink_to(child, target_is_directory=True)
    original_iterdir = Path.iterdir

    def fake_iterdir(path_self: Path):
        if path_self == bad:
            raise OSError("nested read error")
        return original_iterdir(path_self)

    monkeypatch.setattr(Path, "iterdir", fake_iterdir)
    projects = project._discover_legacy_project_directories(root)
    assert child / "project" in projects


def test_load_configuration_delegates(
    tmp_path: Path, monkeypatch: pytest.MonkeyPatch
) -> None:
    cfg = tmp_path / ".kanbus.yml"
    cfg.write_text("x", encoding="utf-8")
    expected = build_project_configuration()
    monkeypatch.setattr(project, "load_project_configuration", lambda _p: expected)
    assert project._load_configuration(cfg) == expected
